/*
 * JobSchedulerFTPSend.java
 * Created on 20.08.2005
 * 
 */
package sos.scheduler.ftp;

import java.io.File;
import java.util.Iterator;
import java.util.Vector;

import sos.net.SOSFTP;
import sos.scheduler.job.JobSchedulerJob;
import sos.util.SOSFile;
import sos.util.SOSSchedulerLogger;

/**
 * 
 *
 * @author Andreas Pschel 
 */
public class JobSchedulerFTPSend extends JobSchedulerJob {

	
	public boolean spooler_process() throws Exception {
	    
		boolean rc = false;
		SOSFTP ftpClient = null;
		
		String host = "";
		int port = 21;
		
		String user = "anonymous";
		String password = "";
		String account = "";
		
		String transferMode = "binary";
		boolean passiveMode = false;
		
		String remoteDir = "/";
		String localDir = ".";
		String fileSpec = ".*";
		
		boolean isLoggedIn = false;
		
		
		this.setLogger(new SOSSchedulerLogger(spooler_log));

		try {   
		    try { // to get the job parameters
			    if (spooler_task.params().var("ftp_host") != null && spooler_task.params().var("ftp_host").length() > 0)
			        host = spooler_task.params().var("ftp_host");
			    if (spooler_task.params().var("ftp_port") != null && spooler_task.params().var("ftp_port").length() > 0)
			        port = Integer.parseInt(spooler_task.params().var("ftp_port"));

			    if (spooler_task.params().var("ftp_user") != null && spooler_task.params().var("ftp_user").length() > 0)
			        user = spooler_task.params().var("ftp_user");
			    if (spooler_task.params().var("ftp_password") != null && spooler_task.params().var("ftp_password").length() > 0)
			        password = spooler_task.params().var("ftp_password");
			    if (spooler_task.params().var("ftp_account") != null && spooler_task.params().var("ftp_account").length() > 0)
			        account = spooler_task.params().var("ftp_account");

			    if (spooler_task.params().var("ftp_transfer_mode") != null && spooler_task.params().var("ftp_transfer_mode").length() > 0)
			        transferMode = spooler_task.params().var("ftp_transfer_mode");
			    if (spooler_task.params().var("ftp_passive_mode") != null && spooler_task.params().var("ftp_passive_mode").length() > 0)
			        passiveMode = (spooler_task.params().var("ftp_passive_mode").equals("1") ? true : false);

			    if (spooler_task.params().var("ftp_remote_dir") != null && spooler_task.params().var("ftp_remote_dir").length() > 0)
			        remoteDir = spooler_task.params().var("ftp_remote_dir");
			    if (spooler_task.params().var("ftp_local_dir") != null && spooler_task.params().var("ftp_local_dir").length() > 0)
			        localDir = spooler_task.params().var("ftp_local_dir");
				
			    if (spooler_task.params().var("ftp_file_spec") != null && spooler_task.params().var("ftp_file_spec").length() > 0)
			        fileSpec = spooler_task.params().var("ftp_file_spec");
			} catch (Exception e) {
				rc = false;
				throw (new Exception("could not process job parameters: " + e.getMessage()));
			}

			
			if (spooler_task.job().order_queue() != null && spooler_task.job().order_queue().length() > 0) {
			    try { // to get the parameters from the order payload
			        sos.spooler.Order order = spooler_task.order();
			        sos.spooler.Variable_set orderData = (sos.spooler.Variable_set) spooler_task.order().payload();

				    if (orderData.var("ftp_host") != null && orderData.var("ftp_host").length() > 0)
				        host = orderData.var("ftp_host");
				    if (orderData.var("ftp_port") != null && orderData.var("ftp_port").length() > 0)
				        port = Integer.parseInt(orderData.var("ftp_port"));

				    if (orderData.var("ftp_user") != null && orderData.var("ftp_user").length() > 0)
				        user = orderData.var("ftp_user");
				    if (orderData.var("ftp_password") != null && orderData.var("ftp_password").length() > 0)
				        password = orderData.var("ftp_password");
				    if (orderData.var("ftp_account") != null && orderData.var("ftp_account").length() > 0)
				        account = orderData.var("ftp_account");

				    if (orderData.var("ftp_transfer_mode") != null && orderData.var("ftp_transfer_mode").length() > 0)
				        transferMode = orderData.var("ftp_transfer_mode");
				    if (orderData.var("ftp_passive_mode") != null && orderData.var("ftp_passive_mode").length() > 0)
				        passiveMode = (orderData.var("ftp_passive_mode").equals("1") ? true : false);

				    if (orderData.var("ftp_remote_dir") != null && orderData.var("ftp_remote_dir").length() > 0)
				        remoteDir = orderData.var("ftp_remote_dir");
				    if (orderData.var("ftp_local_dir") != null && orderData.var("ftp_local_dir").length() > 0)
				        localDir = orderData.var("ftp_local_dir");
				
				    if (orderData.var("ftp_file_spec") != null && orderData.var("ftp_file_spec").length() > 0)
				        fileSpec = orderData.var("ftp_file_spec");
			    } catch (Exception e) {
					rc = false;
					throw (new Exception("could not process order parameters: " + e.getMessage()));
				}
			}
		

			try { // to check parameters
			    if (host == null || host.length() == 0) throw new Exception("no host was given");
			    if (user == null || user.length() == 0) throw new Exception("no user was given");
			    if (fileSpec == null || fileSpec.length() == 0) throw new Exception("no file specification was given");			    
			    // if (remoteDir == null || remoteDir.length() == 0) throw new Exception("no remote directory was given");
			    // if (localDir == null || localDir.length() == 0) throw new Exception("no local directory was given");
			} catch (Exception e) {
				rc = false;
				throw (new Exception("invalid or insufficient parameter: " + e.getMessage()));
			}
			
			
			try { // to process ftp
			    ftpClient = new SOSFTP(host, port);
			    this.getLogger().info("connected to host " + host + ", port " + port);
			    
			    if (account != null && account.length() > 0) {
			        isLoggedIn = ftpClient.login(user, password, account);
			    } else {
			        isLoggedIn = ftpClient.login(user, password);
			    }
			    
			    if (passiveMode) ftpClient.passive();
			    if (transferMode.equalsIgnoreCase("ascii")) {
			        ftpClient.ascii();
			    } else {
			        ftpClient.binary();
			    }
			    
			    ftpClient.cd(remoteDir);
			    ftpClient.cwd(localDir);
			    this.getLogger().info("local directory: " + localDir + ", remote directory: " + remoteDir + ", file specification: " + fileSpec);
			    
			    int count = 0;
        		Vector filelist = SOSFile.getFilelist( new File(localDir).getAbsolutePath(), fileSpec, 0);
        		Iterator iterator = filelist.iterator();
        		while(iterator.hasNext()) {
        		    File transferFile = (File)iterator.next();
        			this.getLogger().info("sending file: " + transferFile.getAbsolutePath());
        			ftpClient.putFile( transferFile.getAbsolutePath(), transferFile.getName() );
        			count++;
        		}
			    
        		switch (count) {
      		  		case 0: 	this.getLogger().warn("no matching files found"); break;
      		  		case 1: 	this.getLogger().info("1 file transferred"); break;
      		  		default: 	this.getLogger().info(count + " files transferred"); break;
        		}
			    
			    return (spooler_task.job().order_queue() == null) ? false : true;			    			    
			    
		    } catch (Exception e) {
				rc = false;
				throw (new Exception("could not process file transfer: " + e.getMessage()));
			} finally {
			    if (ftpClient != null) {
			        if (isLoggedIn) try { ftpClient.logout(); } catch(Exception e) {} // no error handling
			        if (ftpClient.isConnected()) try { ftpClient.disconnect(); } catch(Exception e) {} // no error handling
			    }
			}
			
		} catch (Exception e){
			getLogger().warn ("ftp processing failed: " + e.getMessage());
	        return false;
		}
	}
}
