
package sos.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.jar.JarEntry;
import java.util.jar.JarException;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

public class JarMerge {
	
// Eigenschaften der Kommandozeile
	private boolean check = false;
	private boolean unzip = false;
	private boolean zip = false;	
	private boolean oldOverwrites = false;
	private String source = null;
	private String targetPath = null;
	private String manifestFile = null;
	private boolean standardManifestToCreate = false;
	private String identity = null;
	
	private JarMerge(String[] args) {
		
		this.identity = this.getClass().getName();
		
		try {
			
			this.parseArguments(args);
			
		} catch (Exception e) {
			System.err.println("\n abort: "+e.getMessage());
			this.printUsage();
			System.exit(-1);
		}
	}
	
	private void printUsage() {
		
		String usage = "\n usage:";
		usage += " java "+this.identity+" check PATH";
		usage += "\n";
		usage += "\n        -all jar/zip archives in directory PATH are tested";
		usage += "\n         for duplicate classnames respecting the package structure";
		usage += "\n";
		usage += "\n        java "+this.identity+" [-o] unzip SOURCE [TARGETPATH]";
		usage += "\n";
		usage += "\n        -extract all class files contained in jar/zip archive(s) to subdirectories";
		usage += "\n         relative to package structure.";
		usage += "\n         SOURCE can be a directory or a single archive.";
		usage += "\n         If the directory TARGETPATH is set the structure of subdirectories is created there.";
		usage += "\n         By default the structure of subdirectories is created in current working directory.";
		usage += "\n         If the flag -o is set older class files overwrite newer ones.";
		usage += "\n         By default newer files overwrite older ones.";
		usage += "\n";
		usage += "\n        java "+this.identity+" [-m] zip SOURCEPATH JARFILE [MANIFESTFILE]";
		usage += "\n";
		usage += "\n        -all class files of directory SOURCEPATH and subdirectories are compressed";
		usage += "\n         to jar archive file JARFILE";
		usage += "\n         If a manifest file is set for MANIFESTFILE this file will be added to jar archive.";
		usage += "\n         If the flag -m is set a standard manifest will be created.";
		usage += "\n         This flag will be ignored if MANIFESTFILE is set.";
		usage += "\n         By default it will be searched in SOURCEPATH for '/META-INF/MANIFEST.MF'.";
		usage += "\n         If this manifest file is found it will be added to jar archive.";
		usage += "\n         If this manifest file is not found a standard manifest will be created.";
		usage += "\n";
		usage += "\n        java "+this.identity+" -h";
		usage += "\n";
		usage += "\n        -show this help";
		
		System.out.println(usage);
	}
	
	private void parseArguments(String[] args) throws Exception {
		
		if ( args==null )
			throw new Exception("args is null");
		
		if ( args.length==0 )
			throw new Exception("arguments missing");
		
	// Hilfe anzeigen
		if ( args.length==1 && (args[0].equals("-h") || args[0].equals("-H")) ) {
			this.printUsage();
			System.exit(0);
		}
		
	// search Command
		if ( args[0].equals("check") ) {
			this.check = true;
		}
		else	
		if ( args[0].equals("unzip") ) {
			this.unzip = true;
		}
		else	
		if ( args[0].equals("zip") ) {
				this.zip = true;
		}
		else	
		if ( args[0].equals("-o") || args[0].equals("-O") ) {
			this.oldOverwrites = true;
			if ( args.length<3 ) throw new Exception("arguments missing");
			if ( args.length>4 ) throw new Exception("too much arguments");
			
			if ( args[1].equals("unzip") ) {
				this.unzip = true;
			} else if ( args[1].equals("check") || args[1].equals("zip") ) {
				throw new Exception("unexpected command "+args[1]);
			} else if ( args[1].equals("-m") || args[1].equals("-M") ) {
				throw new Exception("unexpected flag "+args[1]);
			} else
				throw new Exception("unknown command "+args[1]);
		}
		else	
		if ( args[0].equals("-m") || args[0].equals("-M") ) {
			this.standardManifestToCreate = true;
			if ( args.length<4 ) throw new Exception("arguments missing");
			if ( args.length>5 ) throw new Exception("too much arguments");
			
			if ( args[1].equals("zip") ) {
				this.zip = true;
			} else if ( args[1].equals("check") || args[1].equals("unzip") ) {
				throw new Exception("unexpected command "+args[1]);
			} else if ( args[1].equals("-o") || args[1].equals("-O") ) {
				throw new Exception("unexpected flag "+args[1]);
			} else
				throw new Exception("unknown command "+args[1]);
		}
		else
			throw new Exception("unknown command "+args[0]);
		
	// check
		if ( this.check ) {
			if ( args.length<2 ) throw new Exception("arguments missing");
			if ( args.length>2 ) throw new Exception("too much arguments");
			for ( int i=0; i<args.length; i++ ) {
				if ( !args[i].equals("check") ) this.source = args[i];
			}
		}
		
	// unzip
		if ( this.unzip ) {
			for ( int i=0; i<args.length; i++ ) {
				if ( args[i].equals("-o") || args[i].equals("-O") ) this.oldOverwrites = true;
			}
			
			if ( this.oldOverwrites ) {
				if ( args.length<3 ) throw new Exception("arguments missing");
				if ( args.length>4 ) throw new Exception("too much arguments");
			} else {
				if ( args.length<2 ) throw new Exception("arguments missing");
				if ( args.length>3 ) throw new Exception("too much arguments");
			}
			
			for ( int i=0; i<args.length; i++ ) {
				if ( !( args[i].equals("unzip") || args[i].equals("-o") || args[i].equals("-O") ) )
					if ( this.source==null )
						this.source = args[i];
					else
						this.targetPath = args[i];
			}
		}
		
	// zip
		if ( this.zip ) {
			for ( int i=0; i<args.length; i++ ) {
				if ( args[i].equals("-m") || args[i].equals("-M") ) this.standardManifestToCreate = true;
			}
			
			if ( this.standardManifestToCreate ) {
				if ( args.length<4 ) throw new Exception("arguments missing");
				if ( args.length>5 ) throw new Exception("too much arguments");
			} else {
				if ( args.length<3 ) throw new Exception("arguments missing");
				if ( args.length>4 ) throw new Exception("too much arguments");
			}
			
			for ( int i=0; i<args.length; i++ ) {
				if ( !(args[i].equals("zip") || args[i].equals("-m") || args[i].equals("-M")) )
					if ( this.source==null )
						this.source = args[i];
					else {
						if ( this.targetPath==null )
							this.targetPath = args[i];
						else
							this.manifestFile = args[i];
					}
			}
		}
		
	}
	
	/**
	 * Filtert aus einer Liste von Dateinamen alle Dateinamen mit der Endung jar oder zip heraus. 
	 * @return gefilterte Liste
	 */
	private String[] archiveFilter(String[] list) {
		
		String[] fList;
		String suffix;
		int deleted = 0;
		
		if ( list==null ) return null;
		
		for (int i=0; i<list.length; i++) {
			
			if ( list[i].length()<4 || list[i]==null ) {
				list[i] = null;
				deleted++;
			} else {
				suffix = list[i].substring(list[i].length()-4);
				if ( !(suffix.equals(".jar") || suffix.equals(".zip")) ) {
					list[i] = null;
					deleted++;
				}
			}
		}
		
		fList = new String[list.length-deleted];
		
		for (int n=0, i=0; i<list.length; i++) {
			
			if ( list[i]!=null ) fList[n++] = list[i];
		}
		
		return fList;
	}
	
	/**
	 * Filtert aus einer Liste von Dateien alle Dateien mit der Endung class heraus. 
	 * @return gefilterte Liste von Dateien
	 */
	private File[] classFilter(File[] list) {
		
		File[] fList;
		String suffix;
		int deleted = 0;
		
		if ( list==null ) return null;
		
		for (int i=0; i<list.length; i++) {
			
			if ( list[i].getName().length()<6 || list[i]==null ) {
				list[i] = null;
				deleted++;
			} else {
				suffix = list[i].getName().substring(list[i].getName().length()-6);
				if ( !suffix.equals(".class") ) {
					list[i] = null;
					deleted++;
				}
			}
		}
		
		fList = new File[list.length-deleted];
		
		for (int n=0, i=0; i<list.length; i++) {
			
			if ( list[i]!=null ) fList[n++] = list[i];
		}
		
		return fList;
	}
	
	private void executeCheck() throws Exception {
		
		JarFile jar;
		Enumeration content;
		JarEntry entry;
		String classFile;
		long lDate;
		long size;
		
		StringTokenizer st;
		SimpleDateFormat sdf;
		String firstFile;
		String firstDate;
		String firstSize;
		int gap;
		
		File file = new File(this.source);
		
		// alle zu durchsuchenden jar- und zip-Archive ermitteln
		String[] list = this.archiveFilter(file.list());
		
		if ( list.length == 0 ) {
			System.out.println("\n no jar/zip archives found");
			return;
		}
		
		Hashtable inventory = new Hashtable();
		
		try {
			
			// alle Archive durchgehen
			for (int i=0; i<list.length; i++) {
				
				jar = new JarFile(this.source + list[i]);
				
				content = jar.entries();
				
				// alle Klassen eines Archivs durchgehen
				while ( content.hasMoreElements() ) {
					
					entry = (JarEntry)content.nextElement();
					classFile = entry.getName();
					lDate = entry.getTime();
					size = entry.getSize();
					
					if ( classFile.length()>6 )
						if ( classFile.substring(classFile.length()-6).equals(".class") )
							if ( !inventory.containsKey(classFile) ) {
								
								// 								classFile => "x.jar|date|size"
								inventory.put(classFile, list[i]+"|"+String.valueOf(lDate)+"|"+String.valueOf(size));
								
							} else {
								
								st = new StringTokenizer( (String)inventory.get(classFile), "|" );
								sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm");
								
								firstFile = st.nextToken();
								firstDate = st.nextToken();
								firstSize = st.nextToken();
								
								if ( Long.parseLong(firstDate) < 0 )
									firstDate = "unknown";
								else
									firstDate = sdf.format(new Date(Long.parseLong(firstDate)));
								
								if ( Long.parseLong(firstSize) < 0 )
									firstSize = "unknown";
								
								// die grte Stringlnge ermitteln
								gap = Math.max(firstFile.length(), firstDate.length());
								gap = Math.max(gap, firstSize.length());
								
								// und Stringlnge normalisieren
								while ( firstFile.length()<gap ) firstFile += " ";
								while ( firstDate.length()<gap ) firstDate += " ";
								while ( firstSize.length()<gap ) firstSize += " ";
								
								System.out.println("\n duplicate class found: "+classFile);
								System.out.println(" "+ firstFile +"   "+ list[i] );	// Dateiname
								System.out.println(" "+ firstDate +"   "+ ( lDate < 0 ? "unknown" : sdf.format(new Date(lDate)) ) );	// Datum								
								System.out.println(" "+ firstSize +"   "+ ( size < 0 ? "unknown" : String.valueOf(size) ) );	// Dateigre
							}
				}
			}
			
		} catch (Exception e) {
			System.err.println("check error: "+e.getMessage());
		}
	}
	
	private void unpackJarEntry( JarFile jarFile, JarEntry jarEntry, String targetPath) throws JarException, IOException, Exception {
		
		try {
			
			if ( jarFile==null ) throw new Exception("jarFile is null");
			if ( jarEntry==null ) throw new Exception("jarEntry is null");
			if ( targetPath==null ) targetPath = "";
			
			File file = new File( targetPath + jarEntry.getName() );
			
			BufferedInputStream bis = new BufferedInputStream( jarFile.getInputStream( jarEntry ) );
			File dir = new File( file.getParent() );
			dir.mkdirs();
			
			BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( file ) );
			
			for ( int c; ( c = bis.read() ) != -1; )
				bos.write( (byte)c );
			
			bos.close();
			bis.close();
			
			file = new File( targetPath + jarEntry.getName() );
			file.setLastModified( jarEntry.getTime() );
			
		} catch (JarException jax) {
			throw new JarException(jax.getMessage()+": "+this.identity+".unpackJarEntry: "+jarFile.getName() + " => " + targetPath + jarEntry.getName());
		} catch (IOException iox) {
			throw new IOException(iox.getMessage()+": "+this.identity+".unpackJarEntry: "+jarFile.getName() + " => " + targetPath + jarEntry.getName());
		} catch (Exception e) {
			throw new Exception(e.getMessage()+": "+this.identity+".unpackJarEntry: "+jarFile.getName() + " => " + targetPath + jarEntry.getName());
		}
	}
	
	private void createManifestFile(ArrayList iSecName, ArrayList iSecOrigin, ArrayList iSecDate, String targetPath) throws IOException, Exception {
		
		Manifest manifest = null;
		File file = null;
		
		try {
			
			if (iSecName==null || iSecOrigin==null || iSecDate==null)
				throw new Exception("iSec is null");
			if ( !((iSecName.size()==iSecOrigin.size()) && (iSecName.size()==iSecDate.size())) )
				throw new Exception("iSec ArrayList sizes are inconsistent "+iSecName.size()+","+iSecOrigin.size()+","+iSecDate.size());
			
		// Falls keine Klasse entpackt wurde, soll auch kein Manifest geschrieben werden
			if ( iSecName.size()==0 ) return;
			
			if ( targetPath==null ) targetPath = "";
			
			file = new File(targetPath + "META-INF");
			if ( !file.isDirectory() )
				file.mkdir();
			
			file = new File(targetPath + "META-INF/MANIFEST.MF");
			file.createNewFile();
			
			StringBuffer sbuf = new StringBuffer();
			sbuf.append("Manifest-Version: 1.0\n");
			sbuf.append("Created-By: "+System.getProperty("java.vm.version")+" ("+System.getProperty("java.vm.vendor")+")\n");
			sbuf.append("\n");
			
			for (int i=0; i<iSecName.size(); i++) {
				sbuf.append("Name: "+ (String)iSecName.get(i) +"\n");
				sbuf.append("Origin: "+ (String)iSecOrigin.get(i) +"\n");
				sbuf.append("Archive-Date: "+ (String)iSecDate.get(i) +"\n");
				sbuf.append("\n");
			}
			
			// Convert the string to a input stream
			InputStream is = new ByteArrayInputStream(sbuf.toString().getBytes("UTF-8"));
			
			// Create the manifest
			manifest = new Manifest(is);
			manifest.write(new FileOutputStream(file));
			
			System.out.println("\n manifest file "+ file.getCanonicalPath() +" created");
			
		} catch (IOException iox) {
			throw new IOException(this.identity+"constructManifest: "+iox.getMessage());
		} catch (Exception e) {
			throw new Exception(this.identity+"constructManifest: "+e.getMessage());
		}
	}
	
	private Manifest readManifestFile(String path) throws IOException {
		
		Manifest manifest = null;
		
		if ( path==null ) return null;
		
		File file = new File(path);
		
		if ( file.canRead() ) {
			
			try {
				manifest = new Manifest(new FileInputStream(file));
			} catch (FileNotFoundException nfx) {
			} catch (IOException iox) {
				throw new IOException(this.identity+"readManifestFile: "+file.getCanonicalPath()+": "+iox.getMessage());
			}
		}
		
		return manifest;
	}
	
	private void executeUnzip() throws Exception {
		
		File file = null;
		JarFile jar;
		Enumeration content;
		JarEntry entry;
		String classFile = "";
		long archiveDate;
		long hdDate;
		int extracted = 0;
		int processed = 0;
		
		// individuelle Manifest Sektionen
		ArrayList iSecName = new ArrayList();
		ArrayList iSecOrigin = new ArrayList();
		ArrayList iSecDate = new ArrayList();
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm");
		String jarDate;
		int index;
		
		// Verzeichnis der Archive einlesen
		file = new File( this.source );
		
		// alle zu durchsuchenden jar- und zip-Archive ermitteln
		String path = null;
		String[] list = null;
		
		if ( file.isDirectory() ) {
			path = this.source;
			list = this.archiveFilter(file.list());
		} else {
			path = file.getPath().substring(0,file.getPath().length()-file.getName().length());
			list = new String[1];
			list[0] = file.getName();
		}
		
		if ( list.length == 0 ) {
			System.out.println("\n no jar/zip archives found");
			return;
		}
		
		if ( this.oldOverwrites )
			System.out.println("\n directive: overwrite newer versions");
		else
			System.out.println("\n directive: overwrite older versions");
		
		try {
			
			// alle Archive durchgehen
			for (int i=0; i<list.length; i++) {
				
				System.out.println("\n unpack "+ list[i] +":");
				
				jar = new JarFile(path + list[i]);
				content = jar.entries();
				
				file = new File(path + list[i]);
				jarDate = ( file.lastModified() < 0 ) ? "unknown" : sdf.format(new Date(file.lastModified()));  
				
				// alle Klassen eines Archivs durchgehen
				while ( content.hasMoreElements() ) {
					
					entry = (JarEntry)content.nextElement();
					classFile = entry.getName();				
					
					if ( classFile.length()>6 )
						if ( classFile.substring(classFile.length()-6).equals(".class") ) {
							
							file = new File( ((this.targetPath!=null) ? this.targetPath : "") + classFile );
							
							// Datei existiert schon auf Platte
							if ( file.canRead() ) {
								
								hdDate = file.lastModified();
								archiveDate = entry.getTime();
								
								if ( (!this.oldOverwrites && (archiveDate > hdDate)) || (this.oldOverwrites && (archiveDate < hdDate)) ) {
									this.unpackJarEntry(jar, entry, this.targetPath);
									
									// fr Manifest individuelle Sektionen
									if ( iSecName.contains(classFile) ) {
										index = iSecName.indexOf(classFile);
										
										iSecName.remove(index);
										iSecOrigin.remove(index);
										iSecDate.remove(index);
										
										iSecName.add(index, classFile);
										iSecOrigin.add(index, list[i]);
										iSecDate.add(index, jarDate);
										
									} else {
										iSecName.add(classFile);
										iSecOrigin.add(list[i]);
										iSecDate.add(jarDate);
									}
									
									System.out.println("  extract and overwrite "+classFile);
									extracted++;
								} else {
									System.out.println("  do not overwrite      "+classFile);
									processed++;
								}
								
							// Datei existiert noch nicht auf Platte
							} else {
								
								this.unpackJarEntry(jar, entry, this.targetPath);
								
								System.out.println("  extract and write     "+classFile);
								iSecName.add(classFile);
								iSecOrigin.add(list[i]);
								iSecDate.add(jarDate);
								extracted++;
							}
						}
				}
			}
			
			this.createManifestFile(iSecName, iSecOrigin, iSecDate, this.targetPath);
			
			processed += extracted;
			
			System.out.println("\n "+ list.length +" archive"+ ((list.length != 1) ? "s" : "") +" passed");
			System.out.println(" "+ processed +" class"+ ((processed != 1) ? "es" : "") +" passed");
			System.out.println(" "+ extracted +" class"+ ((extracted != 1) ? "es" : "") +" extracted");
			
		} catch (Exception e) {
			System.err.println("unzip error: "+ e.getMessage() +": "+ classFile);
		}
	}
	
	private void createJarArchive(File jarFile, File[] tobeJared, String manifestFile) throws IOException, Exception {
		
		if ( jarFile==null ) throw new Exception("jarFile is null");
		if ( tobeJared==null ) throw new Exception("array of classes is null");
		if ( tobeJared.length==0 ) throw new Exception("no classes to archive");
		
		int BUFFER_SIZE = 10240;
		String path;
		
		try {
			
			byte buffer[] = new byte[BUFFER_SIZE];
			
			// Archiv ffnen
			FileOutputStream stream = new FileOutputStream(jarFile);
			
			Manifest manifest = readManifestFile(manifestFile);
			
			if ( manifest!=null ) {
				File file = new File(manifestFile);
				System.out.println("\n adding existing manifest file "+file.getCanonicalPath());
			
			} else {
				System.out.println("\n create new manifest");
				StringBuffer sbuf = new StringBuffer();
				sbuf.append("Manifest-Version: 1.0\n");
				sbuf.append("Created-By: "+System.getProperty("java.vm.version")+" ("+System.getProperty("java.vm.vendor")+")\n");
				sbuf.append("\n");
				
				// Den String in InputStream berfhren
				InputStream is = new ByteArrayInputStream(sbuf.toString().getBytes("UTF-8"));
				
				// Manifest erzeugen
				manifest = new Manifest(is);
			}
			
			JarOutputStream out = new JarOutputStream(stream, manifest);
			
			JarEntry jarAdd;
			
			for (int i = 0; i < tobeJared.length; i++) {
				
				if ( tobeJared[i] == null )
					continue;
				if ( !tobeJared[i].exists() || tobeJared[i].isDirectory() )
					continue;
				
				// Zielpfad der Datei auf die package-Struktur reduzieren
				path = tobeJared[i].getPath().substring(this.source.length());
				path = path.replaceAll("\\\\", "/");
				
				System.out.print("\n adding " + path);
				
				// Add archive entry
				jarAdd = new JarEntry(path);
				jarAdd.setTime(tobeJared[i].lastModified());
				out.putNextEntry(jarAdd);
				
				// Klasse ins Archiv schreiben
				FileInputStream in = new FileInputStream(tobeJared[i]);
				
				while (true) {
					
					int nRead = in.read(buffer, 0, buffer.length);
					if ( nRead <= 0 )
						break;
					
					out.write(buffer, 0, nRead);
				}
				
				in.close();
			}
			
			out.close();
			stream.close();
			
			System.out.println("\n "+tobeJared.length+" class file"+ ((tobeJared.length != 1) ? "s" : "") +" added");
			
		} catch (IOException iox) {
			throw new IOException(iox.getMessage()+": "+this.identity+".createJarArchive: "+jarFile.getAbsolutePath());
		
		} catch (Exception e) {
			throw new Exception(e.getMessage()+": "+this.identity+".createJarArchive: "+jarFile.getAbsolutePath());
		}
	}
	
	private ArrayList getAllFiles(File file, ArrayList allFiles) {
		
		if ( allFiles==null ) allFiles = new ArrayList();
		
		if ( file.isDirectory() ) {
			
			String[] list = file.list();
			
			for (int i=0; i<list.length; i++)
				this.getAllFiles(new File(file.getPath()+"/"+list[i]), allFiles);
			
		} else {
			allFiles.add(file);
		}
		
		return allFiles;
	}
	
	private void executeZip() throws Exception {
		
		File file = null;
		ArrayList classFiles;
		
		try {
			
		// jar-Archiv berprfen
			file = new File(this.targetPath);
			
			if ( file.isFile() )
				file.delete();
			
			file.createNewFile();
			
		} catch (IOException iox) {
			
			if ( file != null )
				throw new Exception(iox.getMessage()+": "+file.getCanonicalPath());
			else
				throw new Exception(iox.getMessage()+": "+this.targetPath);
		}
		
		File directory = new File(this.source);
		classFiles = new ArrayList();
		classFiles = this.getAllFiles(directory, classFiles);
		
		File[] classes = this.classFilter((File[])classFiles.toArray(new File[classFiles.size()]));
		
		if ( classes.length>0 ) {
			
		// MANIFESTFILE ist gesetzt
			if ( this.manifestFile!=null )
				this.createJarArchive(file, classes, this.manifestFile);
			
		// -m ist gesetzt und MANIFESTFILE ist nicht gesetzt
			else if ( this.standardManifestToCreate )
				this.createJarArchive(file, classes, null);
			
		// default: weder -m noch MANIFESTFILE ist gesetzt
			else
				this.createJarArchive(file, classes, this.source+"META-INF/MANIFEST.MF");
			
			System.out.println("\n "+file.getCanonicalPath()+" created");
			
		} else {
			System.out.println("\n no class files found");
		}
	}
	
	private void execute() {
		
	// Pfad- und Dateiangaben auf Korrektheit berprfen
		try {
			
			File file = new File(this.source);
			
			if ( !file.canRead() )
				throw new Exception(this.source +" not found or not readable");
			
			if ( this.check || this.zip ) {
				if ( !file.isDirectory() )
					throw new Exception(file.getCanonicalPath() +" is no directory");
					
			} else if ( this.unzip ) {
				if ( !file.isDirectory() ) {
					if ( !file.isFile() ) {
						throw new Exception(this.source +" is no normal file");
					} else {
						if ( this.source.length()<5 )
							throw new Exception(file.getCanonicalPath() +" is no valid archiv");
						else {
							String suffix = this.source.substring(this.source.length()-4);
							if ( !(suffix.equals(".jar") || suffix.equals(".zip")) )
								throw new Exception(file.getCanonicalPath() +" is no valid archiv");
						}
					}
				}
			}
			
		// Quellpfad normalisieren
			if ( file.isDirectory() ) {
				this.source = this.source.replaceAll("\\\\", "/");
				if ( !(this.source.charAt(this.source.length()-1)=='/') )
					this.source += "/";
			}
			
		// zustzlich fr unzip optionales Zielverzeichnis testen
			if ( this.unzip ) {
				if ( this.targetPath!=null ) {
					try {
						
						file = new File(this.targetPath);
						
						if ( !file.isDirectory() ) {
							throw new Exception(this.targetPath +" is no directory");
						} else {
							if ( !file.canWrite() )
								throw new Exception(this.targetPath +" is not writable");
						}
						
					// Zielpfad normalisieren
						this.targetPath = this.targetPath.replaceAll("\\\\", "/");
						if ( !(this.targetPath.charAt(this.targetPath.length()-1)=='/') )
							this.targetPath += "/";
						
					} catch (IOException iox) {
						
						if ( file!=null ) {
							if ( file.isFile() )
								throw new Exception(file.getCanonicalPath() +": "+ iox.getMessage());
							else
								throw new Exception(this.targetPath +": "+ iox.getMessage());
						} else
							throw new Exception(this.targetPath +": "+ iox.getMessage());
					}
				}
			}
			
		// zustzlich fr zip Manifest und Ziel-Archiv testen
			if ( this.zip ) {
				
				// Manifest berprfen
				if ( this.manifestFile!=null ) {
					file = new File(this.manifestFile);
					if ( !file.canRead() )
						throw new Exception("manifest file "+file.getCanonicalPath()+" not found or not readable");
				}
				
				// jar-Archiv berprfen
				file = new File(this.targetPath);
				
				if ( file.isDirectory() )
					throw new Exception( file.getCanonicalPath() + " is a directory");
			}
			
		} catch (Exception e) {
			System.err.println("\n abort: "+e.getMessage());
			System.exit(-1);
		}
		
	// spezifizierte Aufgabe ausfhren
		try {
			
			if ( this.check ) this.executeCheck(); else
			if ( this.unzip ) this.executeUnzip(); else
			if ( this.zip   ) this.executeZip();
			
		} catch (Exception e) {
			System.err.println("\n error: "+e.getMessage());
			System.exit(-1);
		}
		
		System.exit(0);
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		JarMerge jarmerge = new JarMerge(args);
		jarmerge.execute();
	}
}
