// $Id: factory_rtf.cxx 3741 2005-06-24 13:22:39Z jz $        Joacim Zschimmer

#include "precomp.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <locale>

#ifdef __GNUC__
#   include <unistd.h>      // unlink
#endif

#include "../kram/sos.h"

#include "../zschimmer/rtf/rtf.h"
#include "../zschimmer/z_com.h"
#include "../zschimmer/z_com_server.h"
#include "../zschimmer/log.h"
#include "../hostole/hostole_version.h"

#include "factory_processor.h"
#include "factory_rtf.h"

using namespace zschimmer;
using namespace zschimmer::rtf;
using namespace ::std;

namespace sos {
namespace factory {

//-------------------------------------------------------------------------------Typbibliothek

DESCRIBE_CLASS( &hostole_typelib, Rtf_constructor, factory_rtf_constructor, CLSID_Factory_rtf_constructor, "DocumentFactoryC.RTF_constructor" , "1.0" );

//-----------------------------------------------------------Rtf_constructor::Rtf_constructor

Rtf_constructor::Rtf_constructor( Site* site, Has_rtf_context* c ) 
: 
    Constructor( factory_rtf_constructor_class_ptr, this ),
    Has_rtf_context( c ),
    _zero_(this+1),
    _allowed_properties( flag_prop | flag_learned_prop ),
    _properties(c)
{
    _site = site;
    initialize();
}

//----------------------------------------------------------Rtf_constructor::~Rtf_constructor

Rtf_constructor::~Rtf_constructor()
{
    for( int i = _templates.size() - 1; i >= 0; i-- )  _templates[i] = NULL;

/*
    if( _work_file_opened )
    {
        _work_file.exceptions( (ios::iostate)0 );
        _work_file.close();
        unlink( _work_filename.c_str() );
    }
*/
}

//------------------------------------------------------------------------Rtf_constructor::_methods
#ifndef SYSTEM_HAS_COM

using namespace zschimmer;
using namespace zschimmer::com;

const Com_method Rtf_constructor::_methods[] =
{ 
    { DISPATCH_METHOD     ,  1, "put"                   , (Com_method_ptr)&Rtf_constructor::Put              , VT_EMPTY, { VT_VARIANT|VT_BYREF, VT_BOOL } },
    { DISPATCH_METHOD     ,  2, "put_formatted"         , (Com_method_ptr)&Rtf_constructor::Put_formatted    , VT_EMPTY, { VT_VARIANT|VT_BYREF, VT_INT } },
    { DISPATCH_METHOD     ,  3, "embed"                 , (Com_method_ptr)&Rtf_constructor::Embed            , VT_EMPTY, { VT_BSTR, VT_INT }, 1 }, 
    { DISPATCH_METHOD     ,  4, "rtf_start"             , (Com_method_ptr)&Rtf_constructor::Rtf_start        }, 
    { DISPATCH_METHOD     ,  5, "rtf_finish"            , (Com_method_ptr)&Rtf_constructor::Rtf_finish       }, 
    { DISPATCH_METHOD     ,  6, "rtf_prop"              , (Com_method_ptr)&Rtf_constructor::Rtf_prop         , VT_EMPTY, { VT_INT, VT_INT } },
    { DISPATCH_METHOD     ,  7, "rtf_put"               , (Com_method_ptr)&Rtf_constructor::Rtf_put          , VT_EMPTY, { VT_INT, VT_INT } },
    { DISPATCH_METHOD     ,  8, "rtf_open"              , (Com_method_ptr)&Rtf_constructor::Rtf_open         , VT_EMPTY, { VT_INT } },
    { DISPATCH_METHOD     ,  9, "rtf_close"             , (Com_method_ptr)&Rtf_constructor::Rtf_close        },
    { DISPATCH_METHOD     , 10, "rtf_include_begin"     , (Com_method_ptr)&Rtf_constructor::Rtf_include_begin, VT_EMPTY },
    { DISPATCH_METHOD     , 11, "rtf_include_end"       , (Com_method_ptr)&Rtf_constructor::Rtf_include_end  , VT_EMPTY },
    { DISPATCH_PROPERTYPUT, 12, "suppress_subdocument"  , (Com_method_ptr)&Rtf_constructor::put_Suppress_subdocument, VT_EMPTY , { VT_BOOL } },
    { DISPATCH_PROPERTYGET, 12, "suppress_subdocument"  , (Com_method_ptr)&Rtf_constructor::get_Suppress_subdocument, VT_BOOL },
    { 0 }
};

#endif
//----------------------------------------------------------------Rtf_constructor::initialize

void Rtf_constructor::initialize()
{
    // Ein gebrauchter Konstruktor kann hiermit erneut verwendet werden.
 
    _doc = Z_NEW( Doc( has_rtf_context(), "Rtf_constructor" ) );
    _doc->_single_lines = _site->_rtf_single_lines;
    _doc->_indent       = _site->_rtf_indent;
    _error_doc = -1;
    _rtf_line_map.clear();
    //_include = false;
}

//------------------------------------------------------------Rtf_constructor::close_doc_file

void Rtf_constructor::close_doc_file()
{
    if( !_work_file.opened() )  return;

    _site->_last_collected_documents_count = _site->_collected_documents_count;
    _site->_collected_documents_count = 0;


    _doc->print_end( &_work_file );

    if( _doc->_doc_properties_changed )
    {
        Z_LOGI2( "rtf", "Rtf_constructor: Der genderte Dokumentkopf wird dem Dokument vorangestellt.\n" );

        _site->_document_copied = true;

        // Dokumentkopf ist erweitert werden (bei zusammengefassten Dokumenten) und muss erneut geschrieben werden.
        Entity_listelem* before_body = _doc->find_last_header_entity();

        z::Stream_file doc_file;
      //doc_file.exceptions( ios::badbit | ios::failbit );
      //doc_file.open( _site->_real_output_filename.c_str(), ios::binary | ios::out | ios::trunc );
        doc_file.open( _site->_real_output_filename.c_str(), "wb" );

        _doc->print_begin( &doc_file );
        _doc->print_entity_list( &doc_file, _doc->_entity.head(), before_body->_next );      // Dokumentkopf

        // Jetzt die gesammelten Dokumentrmpfe aus der Arbeitsdatei kopieren:

        _work_file.seek( _work_file_head_end );
        //_work_file.exceptions( _work_file.exceptions() & ~ios::failbit );

        Dynamic_area buffer ( 1024*1024 );

        while(1)
        {
            int byte_count = _work_file.read( buffer.char_ptr(), buffer.size() );
            if( byte_count == 0 )  break;
            doc_file.write( buffer.char_ptr(), byte_count );
        }

        doc_file.close();
        _work_file.close();
      //_work_file.clear();
        //_work_file_opened = false;
        int ret = unlink( _work_filename.c_str() );    if( ret == -1 )  throw_errno( errno, _work_filename.c_str() );
    }
    else
    {
        _site->_document_copied = false;

        _work_file.close();
        //_work_file_opened = false;
        int ret = rename( _work_filename.c_str(), _site->_real_output_filename.c_str() );   if( ret == -1 )  throw_errno( errno, _site->_real_output_filename.c_str() );
    }

    _site->_document_written = true;

    //!_allowed_properties = flag_prop | flag_learned_prop;        // Wieder alle Eigenschaften zulassen, insbesondere Dokumenteigenschaften

    Z_LOGI2( "rtf", "Rtf_constructor: " << _site->_real_output_filename << " ist fertig.\n" );
}

//---------------------------------------------------------------------Rtf_constructor::close

void Rtf_constructor::close()
{
    if( _work_file.opened() )  close_doc_file();
}

//---------------------------------------------------------------------------Rtf_constructor::start

void Rtf_constructor::start()
{
    if( _doc->first_entity() )  _doc->truncate( _doc->find_last_header_entity() );  // Falls letzte Aufbereitung fehlerhaft war (finish fehlt): Dokumentrumpf lschen

    _doc->_section_break_set = true;        // Folgendes \sbknone unterdrcken (s.a. finish(): \sect\sectd)

    _include           = false;
    _ignore_next_sectd = false;
    _titlepg_inserted  = false;
    _properties        = Properties( has_rtf_context() );      // code_header(f), code_footer(f) zurcksetzen (fr OpenOffice 1.1)
}

//------------------------------------------------------------------Rtf_constructor::rtf_start

STDMETHODIMP Rtf_constructor::Rtf_start()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        start();
    }
    catch( const Xc&        x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_start" ); }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_start" ); }

    return hr;
}

//--------------------------------------------------------------------------Rtf_constructor::finish

void Rtf_constructor::finish()
{
    LOGI2( "rtf", "Rtf_constructor::finish\n" );


    if( _titlepg_inserted )     // Fr OpenOffice 1.1
    {
        // Wenn fr den Briefbogen ein eigener entweder Seitenkopf oder Seitenfu eingerichtet worden ist,
        // also \titlepg eingefgt worden ist, dann muss auch das Gegenteil, Seitenfu oder Seitenkopf,
        // eingefgt werden. Der Seitenfu oder Seitenkopf wird hier auch fr die erste Seite eingerichtet.
        // \footer wird nach \footerf kopiert, bzw. \header nach \headerf.

        Code first_page_code = code_none;    // oder headerf oder footerf 
        Code other_page_code = code_none;    // oder header oder footer

        if( !_properties[ code_headerf ]._entity )  first_page_code = code_headerf,  other_page_code = code_header;
        if( !_properties[ code_footerf ]._entity )  first_page_code = code_footerf,  other_page_code = code_footer;

        if( first_page_code )
        {
            Simple_entity* e = _doc->first_entity();
            if( e )
            {
                while( e->next()  &&  e->next()->code() != other_page_code )  e = e->next();
                if( e->next() )
                {
                    Simple_entity* new_first_page = new_entity( *e->next() );
                    new_first_page->change_code( first_page_code );
                    e->insert_successor( new_first_page );
                }
            }
        }
    }


    //if( _site->_collected_documents_count == 0 )
    if( !_work_file.opened() )
    {
        //time_t t = time(NULL);
        if( _allowed_properties & flag_prop_doc )
        {
            _doc->prepend_info_doccomm( "Document Processor " VER_PRODUCTVERSION_STR "\r\n" );  // "generiert am " + string( asctime( localtime( &t ) ) ) + "\r\n" );
        }


        if( _site->_output_filename == "" )  throw_xc( "FACTORY-307" );

        _site->_real_output_filename = _site->_output_filename;

        if( _site->_collect )
        {
            _site->_document_nr++;

            // In "###" die Dokumentnummer einsetzen:
            int nr = _site->_document_nr;
            for( int i = _site->_real_output_filename.length() - 1; i >= 0; i-- )
            {
                char c = _site->_real_output_filename[i];
                if( c == '/' )  break;      // '#' im Verzeichnisnamen erstmal nicht ersetzen
                if( c == '\\' )  break;
                if( c == '#' )  _site->_real_output_filename[i] = '0' + ( nr % 10 ),  nr /= 10;
            }
            if( nr  &&  _site->_document_nr > 1 )  throw_xc( "FACTORY-304", _site->_real_output_filename.c_str(), _site->_document_nr );
        }

        _work_filename = _site->_real_output_filename + "~";

        int ret = unlink( _site->_real_output_filename.c_str() );   if( ret == -1  &&  errno != ENOENT )  throw_errno( errno, _work_filename.c_str(), "unlink" );

        //_work_file.exceptions( ios::badbit | ios::failbit );
        _work_file.unlink_later();
        //_work_file.open( _work_filename, ios::binary | ios::in | ios::out | ios::trunc );
        _work_file.open( _work_filename, "w+b" );

        //_work_file_opened = true;

        _doc->print_begin( &_work_file );
        _doc->print_entity_list( &_work_file, _doc->_entity.head(), _doc->find_last_header_entity()->_next );      // Dokumentkopf
        Z_DEBUG_ONLY( _work_file.flush() );

        _work_file_head_end = _work_file.tell();
    }
    else
    {
        if( !_last_subdocument_suppressed )
        {
            _work_file.print( "\n\n\\sect\\sectd\n\n" );      // s.a. rtf_start(): _doc->_section_break_set = true
            Z_DEBUG_ONLY( _work_file.flush() );
        }

        _last_subdocument_suppressed = false;
    }


    
    Entity_listelem* before_body = _doc->find_last_header_entity();
    //Z_LOG2( "joacim", "before_body=" << descr( before_body )->_name << "\n" );
    //Z_LOG2( "joacim", "before_body next=" << descr( before_body->_next )->_name << "\n" );


    if( _suppress_subdocument )
    {
        _suppress_subdocument = false;
        _last_subdocument_suppressed = true;            // Nchstes Subdokument soll nicht mit \sect\sectd abgetrennt werden
    }
    else
    {
        _site->_collected_documents_count++;

        _doc->optimize_body();
        _doc->modify_for_microsoft_word();
        _doc->print_entity_list( &_work_file, before_body->_next, NULL );
        Z_DEBUG_ONLY( _work_file.flush() );
    }

    if( _site->_collected_documents_count >= _site->_collect  &&  !_site->_different_templates_allowed )  close_doc_file();

    _doc->truncate( before_body );                            // Dokumentrumpf lschen, Dokumentkopf stehen lassen

    _allowed_properties &= ~( flag_prop_doc | flag_learned_prop_doc );  // Beim evtl. nchsten Lauf keine Dokumenteigenschaften bernehmen!

    LOG2( "joacim", "Rtf_constructor::finish ok\n" );
}

//------------------------------------------------------------------Rtf_constructor::rtf_finish

STDMETHODIMP Rtf_constructor::Rtf_finish()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        finish();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_finish" ); }

    return hr;
}

//------------------------------------------------------------------Rtf_constructor::put_string

void Rtf_constructor::put_string( const string& text, bool )
{
    const char* p       = text.c_str();
    const char* p_end   = p + text.length();

    if( p_end - p > 6  &&  memcmp( p, "{\\rtf1", 6 ) == 0  &&  p_end[-1] == '}' )     // RTF-Ausdruck?
    {
        Embed_synthesizer   synth  ( _site, this );
        Rtf_parser          parser ( has_rtf_context() );

        parser._site                = _site;
        parser._synth               = &synth;
        parser.set_constructor      ( this );
        parser._include             = true;
        parser._dont_merge          = true;   // RTF-Variable hat keine \fonttbl etc., alle Nummern sind schon korrekt
        parser._allowed_properties  = flag_prop_char;
        parser._restore_properties  = true;
        parser.init();
        parser.parse( text );
    }
    else
    {
        while( *p ) 
        {
            const char* p0 = p;
            while( (unsigned char)p[0] >= ' ' )  p++;

            if( p > p0 )
            {
                Entity entity ( has_rtf_context(), code_text );
                entity.set_text( string( p0, p - p0 ) );
                _doc->append_entity( entity );
            }

            if( *p )
            {
                switch( p[0] )
                {
                    case '\n': _doc->append_entity( code_line );  break;
                    case '\f': _doc->append_entity( code_page );  break;
                    default  : break;   // nicht beachten
                }

                p++;
            }
        }
    }
}

//----------------------------------------------------------Rtf_constructor::handle_inserted_entity

void Rtf_constructor::handle_inserted_entity( Simple_entity* entity )
{
    if( !_doc->is_in_destination() )
    {
        if( _include )
        {
            // Fr OpenOffice 1.1

            switch( entity->code() )
            {
                case code_header:
                case code_footer:
                {
                    Code first_page_code = entity->code() == code_header? code_headerf : code_footerf;  // code_headerf oder code_footerf
                    Code other_page_code = entity->code();                                              // code_header  oder code_footer

                    // Ein \header in einem #include macht den bereits aus der umgebenden Vorlage erzeugten \header zu \headerf, also nur fr die erste Seite.

                    if( _properties[ first_page_code ]._entity == NULL 
                    &&  _properties[ other_page_code ]._entity != NULL )
                    {
                        // \header des Briefbogens zu \headerf machen
                        // \header der Vorlage hinter \headerf schieben
                        // \titlepg voranstellen
                        Simple_entity* e;

                        Destination_entity* outer = cast_destination_entity( _properties[ other_page_code ]._entity );
                        outer->change_code( first_page_code );
                        _properties[ first_page_code  ] = _properties[ other_page_code ];

                        if( !_properties[ code_titlepg ]._value )
                        {
                            e = _doc->first_entity();
                            while( e  &&  e->next() != outer )  e = e->next();
                            if( e )
                            {
                                Simple_entity* titlepg = new Simple_entity( code_titlepg );
                                e->insert_successor( titlepg );
                                _properties.set_property( titlepg );
                                _titlepg_inserted = true;
                            }
                        }

                        e = new_entity( *entity );
                        outer->insert_successor( e );

                        _properties[ other_page_code ]._entity = e;

                        while( e  &&  e->next() != entity )  e = e->next();
                        if( e )  _doc->truncate( e );
                    }

                    break;
                }

                default:
                    break;
            }
        }
        else
        {
            if( entity->is_property() )  _properties.set_property( entity );
        }
    }
}

//-------------------------------------------------------------------Rtf_constructor::append_entity

Entity_listelem* Rtf_constructor::append_entity( const Simple_entity& entity )
{
    if( descr( entity.code() )->is_non_allowed_property( _allowed_properties ) )  return NULL;

    switch( entity.code() )
    {
        case code_sectd:
        {
            if( _ignore_next_sectd )
            {
                _ignore_next_sectd = false;
                return NULL;
            }

          //if( !_sect_encountered )  
          //    return NULL;

          //_sect_encountered = false;
            break;
        }

        case code_sect:
          //_sect_encountered = true;
            _ignore_next_sectd = false;
            break;

        default: ;
    }

    return _doc->append_entity( entity );  
}

//------------------------------------------------------------------------Rtf_constructor::rtf_prop

void Rtf_constructor::rtf_prop( int c, int value )
{
    Code code = (Code)c;

    append_entity( Entity( has_rtf_context(), code, value ) );
}

//------------------------------------------------------------------------Rtf_constructor::rtf_prop

STDMETHODIMP Rtf_constructor::Rtf_prop( int code, int value )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        rtf_prop( (Code)code, value );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_prop" ); }

    return hr;
}

//-------------------------------------------------------------------------Rtf_constructor::rtf_put

void Rtf_constructor::rtf_put( int begin_ptr, int end_ptr )
{
    for( Entity_listelem* entity = (Entity_listelem*)begin_ptr; entity != (Entity_listelem*)end_ptr;  entity = entity->_next )
    {
        Simple_entity* inserted_entity = append_entity( *entity );
        if( inserted_entity )
        {
            handle_inserted_entity( inserted_entity );
        }
    }
}

//-------------------------------------------------------------------------Rtf_constructor::Rtf_put

STDMETHODIMP Rtf_constructor::Rtf_put( int begin_ptr, int end_ptr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        rtf_put( begin_ptr, end_ptr );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_put" ); }

    return hr;
}

//------------------------------------------------------------------------Rtf_constructor::rtf_open

void Rtf_constructor::rtf_open( int entity_ptr )
{
    Destination_entity* entity = cast_destination_entity( (Simple_entity*)entity_ptr );

    Entity kopf ( has_rtf_context() );
    kopf.shallow_copy( *entity );
    Destination_entity* e = cast_destination_entity( append_entity( kopf ) );
    if( !e )  throw_xc( "Rtf_constructor::rtf_open", descr( kopf.code() )->_name );     // Falls append_entity() den Code verweigert

    _doc->push_current_destination( entity->has_own_properties() );
    _doc->_current_destination = e;
}

//------------------------------------------------------------------Rtf_constructor::rtf_open

STDMETHODIMP Rtf_constructor::Rtf_open( int entity_ptr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        rtf_open( entity_ptr );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_open" ); }

    return hr;
}

//-----------------------------------------------------------------------Rtf_constructor::rtf_close

void Rtf_constructor::rtf_close()
{
    Destination_entity* destination_entity = _doc->_current_destination;

    _doc->pop_current_destination();

    handle_inserted_entity( destination_entity );
}

//------------------------------------------------------------------Rtf_constructor::rtf_close

STDMETHODIMP Rtf_constructor::Rtf_close()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        rtf_close();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_close" ); }

    return hr;
}

//---------------------------------------------------------------Rtf_constructor::rtf_include_begin

STDMETHODIMP Rtf_constructor::Rtf_include_begin()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        _include++;
        _ignore_next_sectd = true;      // Fr OpenOffice 1.1, wenn Briefbogen einen Abschnitt (Spaltendruck) hat. \sectd unterdrckt zweite Kopfzeilen \header. 20.1.2004
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_include_begin" ); }

    return hr;
}

//-----------------------------------------------------------------Rtf_constructor::rtf_include_end

STDMETHODIMP Rtf_constructor::Rtf_include_end()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        _include--;
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Rtf_constructor::rtf_include_end" ); }

    return hr;
}

//-------------------------------------------------------------------Rtf_constructor::map_next_line

void Rtf_constructor::map_next_line( int doc_no, int line_nr, Simple_entity* first, const Entity_ptr& start, const Entity_ptr& end, bool in_script, const string& line )
{
    Z_LOG2( "rtf.map_next_line", "map_next_line(" << doc_no << "," << (line_nr+1) << "): " << line << "\n" );

    Doc_entity doc_entity ( has_rtf_context() );

    doc_entity._doc_no    = doc_no;
    doc_entity._first     = first;
    doc_entity._start     = start;
    doc_entity._end       = end;
    doc_entity._in_script = in_script;

    _rtf_line_map.insert( Rtf_line_map::value_type( line_nr, doc_entity ) );
}

//------------------------------------------------------------------Rtf_constructor::map_next_lines

void Rtf_constructor::map_next_lines( int doc_no, int line_nr_, Simple_entity* first, const Entity_ptr& start, const Entity_ptr& end, bool in_script )
{
    int        line_nr    = line_nr_;
    Entity_ptr p          = start;
    Entity_ptr line_start = start;

    while( p != end )
    {
        if( !p ) 
        { 
            LOG( "*** Rtf_contructor::map_next_lines(): Ende nicht erreicht ***\n" );  
            break; 
        }

        string line;

        while(1)
        {
            int c = p.peek_char_until(end);
            if( c == EOF  )  break;
            if( c != '\n' )  break;
            p.skip_char();
            line_nr++;
        }

        if( p == end )  break;

        while( p != end )
        {
            int c = p.peek_char_until(end);
            if( c == EOF  )  break;
            if( c == '\n' )  break;
            p.skip_char();
            Z_DEBUG_ONLY( line += (char)c );
        }

        map_next_line( doc_no, line_nr, first, line_start, p, in_script, line );

        line_start = p;
    }
}

//----------------------------------------------------------------Rtf_constructor::set_error

void Rtf_constructor::set_error( const Xc& x )
{
    const string style_name    = "SCRIPTERROR";
    const string bookmark_name = "SCRIPTERROR";

    Rtf_line_map::iterator it = _rtf_line_map.find( x._pos._line );
    if( it == _rtf_line_map.end() ) 
    {
        LOG( "Rtf_constructor::set_error: Zeile " << (x._pos._line+1) << " ist unbekannt.\n" );
        return;
    }

    it->second._start.split_text();
    it->second._end.split_text();
    
    _error_doc = it->second._doc_no;

    Doc* doc = _templates[ it->second._doc_no ];


    // Style fr fehlerhafte Zeile:
    ptr<zschimmer::rtf::Doc::Style> style = doc->_header.style_by_name( style_name );
    if( style && style->code() != code_cs )
    {
        doc->remove_style( style );
        style = NULL;
    }

    if( !style )
    {
        style = Z_NEW( Doc::Style( has_rtf_context(), code_cs, style_name ) );
        doc->add_style( style );
        
        style->set_property( code_cf, doc->color( 255, 0, 0 )->number() );
        
        Doc::Style* standard = doc->_header.style_by_name( "Default Paragraph Font" );
        if( standard )
        {
            style->set_property( code_additive );
            style->set_property( code_sbasedon, standard->code() );
        }
    }


/*
    // Altes Lesezeichen lschen:
    for( Simple_entity::Iterator e = doc->_entity.head(); e; )
    {
        if( ( e->code() == code_bkmkstart || e->code() == code_bkmkend )
            && stricmp( c_str(zschimmer::rtf::read_text(e)), c_str(bookmark_name) ) == 0 )
        {
            e->remove_successor();
        }
        else
            e++;
    }
*/

    // Vorangehende Entitt der fehlerhaften Stelle suchen:
    Entity_ptr e            ( has_rtf_context(), it->second._first );   //doc->_entity.head();
    Properties props        ( has_rtf_context() );                      // Eigenschaften am Ende der fehlerhaften Zeile
    Properties style_props  ( has_rtf_context() );                      // Wie props, aber mit Eigenschaften der Zeichenformatvorlage style

    while( e->_next  &&  e->_next != it->second._start ) 
    {
        if( e->is_property() )  props.set_property( e );
        e = e->_next;
    }

    // {\bkmkstart SCRIPT ERROR}
    Destination_entity* bkmkstart = new Destination_entity( has_rtf_context(), code_bkmkstart );
  //bkmkstart->_head = new Text_entity( bookmark_name );
    bkmkstart->append_to_destination( new Text_entity( has_rtf_context(), bookmark_name ) );
    e->insert_successor( bkmkstart );
    e = e->_next;   


    // Zeichenformatvorlage zur Markierung der fehlerhaften Stelle:
    style_props = props;
    if( style ) 
    {
        style->insert_into_doc( e );

        while( e->_next  &&  e->_next->is_property() ) 
        { 
            style_props.set_property( e->_next );  
            e = e->_next; 
        }
    }

    e->insert_successor( new Param_entity( code_b, 1 ) );   // Fett ein
    e = e->_next;

    while( e->_next  &&  e->_next != it->second._end ) 
    {
        if( e->_next->is_property() ) 
        {
            props.set_property( e->_next );
            style_props.set_property( e );
        }

        e = e->_next;
        
        if( e->code() == code_cs )  style = NULL;
        else
        if( e->code() == code_plain )  style = NULL;
    }

    e->insert_successor( new Param_entity( code_b, 0 ) );   // Fett aus
    e = e->_next;

    // Fehlermeldung einfgen:
    string text = x.what();  int i = -1;
    while( ++i < text.length() )  if( text[i] == '\n' )  text.insert( i, _site->_comment ), i+= _site->_comment.length();
    e->insert_successor( new Text_entity( has_rtf_context(), "  " + ( it->second._in_script? _site->_comment : "" ) + "<-- " + text ) ); 
    e = e->_next;


    // {\bkmkend SCRIPT ERROR}
    Destination_entity* bkmkend = new Destination_entity( has_rtf_context(), code_bkmkend );
  //bkmkend->_head = new Text_entity( bookmark_name );
    bkmkend->append_to_destination( new Text_entity( has_rtf_context(), bookmark_name ) );
    e->insert_successor( bkmkend );
    e = e->_next;


    // Vorherige Eigenschaften einstellen (also ohne Zeichenformatvorlage):
    if( style )  props.insert( e, style_props );
}

//----------------------------------------------------------Rtf_constructor::error_filename

string Rtf_constructor::error_filename()
{
    if( _error_doc < 0 )  return "";//throw_xc( "FACTORY-116" );

    return _template_names[ _error_doc ];
}

//----------------------------------------------------------Rtf_constructor::error_document

string Rtf_constructor::error_document()
{
    if( _error_doc < 0 )  return "";//throw_xc( "FACTORY-116" );

    ostringstream s;

    s << *_templates[ _error_doc ];

    return s.str();
}

//------------------------------------------------------------------------------operator<< (Symbol)

ostream& operator<< ( ostream& s, Rtf_parser::Symbol symbol )
{
    switch( symbol )
    {
        case Rtf_parser::sym_none:         s << "sym_none";             break;
        case Rtf_parser::sym_eof:          s << "sym_eof";              break;
        case Rtf_parser::sym_script:       s << "sym_script";           break;
        case Rtf_parser::sym_preprocessor: s << "sym_preprocessor";     break;
        case Rtf_parser::sym_expr:         s << "sym_expr";             break;
        case Rtf_parser::sym_text:         s << "sym_text";             break;
        default:                           s << "sym_" << (int)symbol;
    }

    return s;
}

//-------------------------------------------------------------------------------operator<< (Token)

ostream& operator<< ( ostream& s, const Rtf_parser::Token& token )
{
    s << token._symbol << ":" << token._text;
    return s;
}

//--------------------------------------------------------------------Rtf_parser::Rtf_parser

Rtf_parser::Rtf_parser( Has_rtf_context* c )
: 
    Has_rtf_context(c),
    _zero_(this+1),
    _entity_ptr( has_rtf_context() ),
    _line_entity_ptr( has_rtf_context() ),
    _text_start( has_rtf_context() ),
    _script_end( has_rtf_context() )
{    
    obj_const_name( "factory::Rtf_parser" );
}

//-------------------------------------------------------------------Rtf_parser::~Rtf_parser

Rtf_parser::~Rtf_parser()
{    
}
    
//--------------------------------------------------------------------------Rtf_parser::init

void Rtf_parser::init()
{   
    // Verschiedene Flle beachten:
    // #include (whrend der Generierung des Skripts, make_script)
    // embed() (whrend der Konstruktion, Ausfhrung des Skripts)
    // Serienaufbereitung
    // _different_templates_allowed
    // &("{\rtf1...}")

    if( _constructor )
    {
        // Der Parser ist von #include oder embed() gerufen und soll den
        // bereits vorhandenen Konstruktor (Text_constructor oder Rtf_constructor) benutzen.
        // Oder make_script() wird ein zweites Mal gerufen bei _different_templates_allowed.

        if( typeid( *_constructor ) != typeid(Rtf_constructor) )  throw_xc( "FACTORY-211" );

        _rtf_constructor = static_cast<Rtf_constructor*>( +_constructor );

        if( !_dont_merge  &&  _rtf_constructor->_doc->filled() )  
        //if( !_dont_merge  &&  !_main_doc )
        {
            // Wenn das Ausgabedokument bekonnen wurde, ist das unser Ziel fr die eingemischten Dokumenteigenschaften.
            // Bei Serienaufbereitung, _different_templates_allowed
            _main_doc = _rtf_constructor->_doc;
        }
    }
    else
    {
        _rtf_constructor = new Rtf_constructor( _site, has_rtf_context() );
        set_constructor( _rtf_constructor );
    }

    if( !_dont_merge  &&  !_main_doc  &&  _main_parser )  // #include
    {
        if( typeid(*_main_parser) != typeid(Rtf_parser) )  throw_xc( "FACTORY-211" );

        _main_doc = static_cast<Rtf_parser*>( _main_parser )->_template;  // Hauptdokumentvorlage
    }

    if( _template )
    {
        // _template != NULL nur, wenn RTF-Destination mit Text einen verschachtelten Parser erfordert, s. rtf_put_2()
    }
    else
    {
        _template = Z_NEW( Doc( has_rtf_context(), "template " + _document_name ) );

        if( _rtf_constructor ) 
        {
            _template_index = _rtf_constructor->_templates.size();
            _rtf_constructor->_template_names.push_back( _document_name );
            _rtf_constructor->_templates.push_back( _template );  // Fr ein Leben nach dem Parser: Im Konstruktor
        }
    }


    assert( _allowed_properties );
    //if( _allowed_properties == 0 )  _allowed_properties = (Embed_flags)-1;  // Alles durchlassen
    if( _allowed_properties & flag_prop_para )  _allowed_properties |= flag_prop_para_complete | flag_prop_row | flag_prop_cell;

    //_allowed_properties |= ~flag_prop;    // Nur Eigenschaften filtern, Rest ist sowieso erlaubt
    
    if( _allowed_properties & flag_prop_doc )  _allowed_properties |= flag_learned_prop_doc;
                                         else  _allowed_properties &= ~flag_learned_prop_doc;

/*
    if( _constructor )
    {
        // Der Parser ist von #include oder embed() gerufen und soll den
        // bereits vorhandenen Konstruktor (Text_constructor oder Rtf_constructor) benutzen.
        // Oder make_script() wird mehrfach hintereinander gerufen.

        if( typeid( *_constructor ) == typeid(Rtf_constructor) ) 
        {
            _rtf_constructor = static_cast<Rtf_constructor*>( +_constructor );

            //if( !_include )
            {
                _main_doc = _rtf_constructor->_doc;  //jz 24.10.04  _different_templates_allowed = true  und  zweite Vorlage (deren Kopf mit der ersten Vorlage vermischt werden muss)
                if( _allowed_properties == 0 )  _allowed_properties = (Embed_flags) ~flag_prop_doc;
            }
        }
        else
            throw_xc( "FACTORY-211" );
    }
    else
    {
        _rtf_constructor = new Rtf_constructor( _site, has_rtf_context() );
        set_constructor( _rtf_constructor );
    }


    if( _template )
    {
        // _template != NULL nur, wenn RTF-Destination mit Text einen verschachtelten Parser erfordert, s. rtf_put_2()
    }
    else
    {
        _template = Z_NEW( Doc( has_rtf_context(), "template " + _document_name ) );

        if( _rtf_constructor ) {
            _template_index = _rtf_constructor->_templates.size();
            _rtf_constructor->_template_names.push_back( _document_name );
            _rtf_constructor->_templates.push_back( _template );  // Fr ein Leben nach dem Parser: Im Konstruktor
        }
    }


    if( _include )
    {
        if( !_main_doc )
        {
            if( _main_parser  &&  typeid(*_main_parser) == typeid(Rtf_parser) )
            {
                // #include
                Rtf_parser* main_parser = (Rtf_parser*)_main_parser;
                _main_doc = main_parser->_template;  // Hauptdokumentvorlage
            }
            else
            {
                // embed()
                _main_doc = _rtf_constructor->_doc;  // Dokument, das konstruiert wird   
            }
        }

        if( _allowed_properties == 0 )  _allowed_properties = (Embed_flags) ~(flag_prop_doc|flag_prop_sect);
    }


    if( _allowed_properties == 0 )  _allowed_properties = (Embed_flags)-1;  // Alles durchlassen
    if( _allowed_properties & flag_prop_para )  _allowed_properties = Embed_flags( _allowed_properties | flag_prop_para_complete | flag_prop_row | flag_prop_cell );
    _allowed_properties = (Embed_flags)( _allowed_properties | ~flag_prop );    // Nur Eigenschaften filtern, Rest ist sowieso erlaubt
*/
}
    
//-----------------------------------------------------------------------------Rtf_parser::eat_char

int Rtf_parser::eat_char()
{
    int result = _char;

    if( _char != EOF )
    {
        _eaten_char_pos._line = _char_pos._line;
        _eaten_char_pos._col  = _char_pos._col;
   
        if( _char != '\0' )  _entity_ptr.skip_char();   // Das erste Zeichen nicht berspringen
        _char = _entity_ptr.peek_char();

        if( _char == '\n' )  _char_pos._line++,  _char_pos._col = 0;
                       else  _char_pos._col++;
    }

    return result;
}

//-----------------------------------------------------------------------Rtf_parser::eat_rtf_string

string Rtf_parser::eat_rtf_string()
{
    ostringstream s;

    s << "{\\rtf1";


    // Erstmal die gerade gltigen Zeicheneigenschaften bernehmen

    for( Code prop = code_none; prop < code__properties_max; prop++ )
    {
        if( is_char_property( prop ) 
         && !_entity_ptr._properties.is_default(prop) )
        {
            Param_entity entity ( prop, _entity_ptr._properties[prop] );
            print_entity( &s, entity );
        }
    }
    

    _entity_ptr.split_text();
    Entity_ptr e = _entity_ptr;     // Anfang des RTF-Textes

    eat_string();

    _entity_ptr.split_text();       // Ende des RTF-Textes


    // RTF-Text bernehmen. Mit Haut und Haaren: Funoten, Textmarken, ...

    while( (Simple_entity*)e != (Simple_entity*)_entity_ptr )
    {
        print_entity( &s, *e );
        e.to_next_entity();
    }

    s << "}";

    return quoted_string( c_str(s.str()), _site->_quote_char, _site->_use_backslash? '\\' : _site->_quote_char );
}

//---------------------------------------------------------------------Rtf_parser::eat_string

string Rtf_parser::eat_string()
{
    Source_pos  begin_position  = _token._pos;
    char        quote           = _char;
    string      str             = as_string( (char)eat_char() );

    while(1)
    {
        if( _char == EOF )
        {
            string part = quote + str.substr( 0, 50 );
            if( str.length() > 50 )  part += "...";
            throw_syntax_error( "FACTORY-208", part.c_str(), begin_position );
        }
        
        if( _char == '\\'  &&  _site->_use_backslash ) 
        {
            str += (char)eat_char();
        }
        else
        if( _char == quote  ) 
        { 
            str += eat_char();  
            if( _site->_use_backslash )  break;
            if( _char != quote )  break;
        }

        str += (char)eat_char();
    }
    
    return str;
}

//-----------------------------------------------------------------Rtf_parser::in_script_style

bool Rtf_parser::in_script_style()
{
    _in_script_style = _entity_ptr._properties[ code_s  ] == _script_para_style_no
                    || _entity_ptr._properties[ code_cs ] == _script_char_style_no;
    return _in_script_style;
}

//-----------------------------------------------------------------------Rtf_parser::eat_token

void Rtf_parser::eat_token()
{
    _token._pos    = _char_pos;
    _token._text   = "";
    _token._format = format_std;

    if( !_in_script && _token._symbol != sym_text && !_in_script_style )        // Damit bei ...%>\b<% nix verloren geht
    {
        _token._symbol = sym_text;
        return;
    }

    if( _char == EOF ) 
    {
        _token._symbol = sym_eof;
    }
    else
    if( !_in_script )
    {
        if( in_script_style() )  goto SCRIPT;

        if( _char == _site->_var_char )
        {
            eat_char();
/*falsch                         
            if( _char == _site->_var_char  &&  !_site->_use_backslash ) 
            {
                eat_char();
                _token._symbol = sym_text;
                _token._text = "&"; falsch   _entity_ptr muss verschoben werden
                break;
            }
*/
            if( _char == '$' )  _token._format = format_waehrung,  eat_char();
        
            if( _char == '(' )
            {
                eat_char();
                
                _entity_ptr.split_text();
                _line_entity_ptr = _entity_ptr;

                int klammern = 1;
            
                while(1)
                {
                    if( _char == EOF )  throw_syntax_error( "FACTORY-210", _token._pos );
                    else
                    if( _char == '"'  ||  _char == '\''  ||  _char == '`' )  _token._text += eat_string();  
                    else
                    {
                        if( _char == '(' )  klammern++; 

                        if( _char == ')' ) 
                        {
                            if( klammern == 1 ) {
                                //_entity_ptr.split_text();
                                _script_end = _entity_ptr;
                                _text_start = _entity_ptr,  _text_start.skip_char();
                                eat_char();
                                break;
                            }
                            klammern--;
                        }

                        _token._text += (char)eat_char();
                    }

                    if( klammern == 0 )  break;
                }

                _token._symbol = sym_expr;
            }
            else
            {
                if( isalpha( _char ) )
                {
                    _entity_ptr.split_text();
                    _line_entity_ptr = _entity_ptr;

                    if( _site->_var_prefix )  _token._text += _site->_var_prefix;       // '$' fr Perl

                    while( isalnum( _char )  ||  _char == '_' ) {
                        _text_start = _entity_ptr,  _text_start.skip_char();
                        _token._text += (char)eat_char();
                    }

                    _token._symbol = sym_expr;
                    _script_end = _entity_ptr;
                }
                else
                {
                    _token._text = _site->_var_char;
                    goto TEXT;
                }
            }
        }
        else
        if( _char == '<' )
        {
            eat_char();

            if( _char == '%' )
            {
                _in_script = true;
                eat_char();
                goto SCRIPT;
            }
            else
            {
                _token._text = "<";
                goto TEXT;
            }
        }
        else
        {
TEXT:       _token._symbol = sym_text;
            while( _char != EOF  &&  _char != '<'  &&  _char != _site->_var_char  &&  !in_script_style() )  Z_DEBUG_ONLY( _token._text += )  eat_char();
        }
    }
    else
    {
SCRIPT:
        _token._symbol = sym_script;
        _entity_ptr.split_text();
        _line_entity_ptr = _entity_ptr;
    
        if( _char == '#' )  _token._symbol = sym_preprocessor,  eat_char();

        while(1)
        {
            if( _in_script_style  &&  !in_script_style() ) 
            {
                _text_start = _entity_ptr;
                break;
            }

            switch( _char )
            {
                case EOF :  goto SCRIPT_OK;


                case '"':   // Besonderheit rtf"string" lsst string im RTF-Code durch
                            {
                                int n = _token._text.length();
                                if( n >= 3
                                 && strcmp( _token._text.c_str() + n - 3, "rtf" ) == 0 )
                                {
                                    _token._text.erase( n-3 );
                                    _token._text += eat_rtf_string();
                                    break;
                                }
                            }
                case '\'':  
                case '`':   if( _site->_quotes.find( _char ) != string::npos )  _token._text += eat_string();
                                                                          else  _token._text += (char)eat_char();
                            break;


                case '\r':  eat_char();  break;


                case '\f':  // \page
                case '\n':  // \sect, \par, \line
                            if( _in_script_style ) 
                            {
                                Entity_ptr_with_prop e = _entity_ptr;
                                eat_char();

                                if( !in_script_style() )   // Beginnt Text?
                                { 
                                    _text_start = e; 
                                    _text_start.skip_char(); 
                                }
                            }
                            else
                                eat_char();  

                            goto SCRIPT_OK;


                case '%' :  //_entity_ptr.split_text();
                            _script_end = _entity_ptr;
                            eat_char();

                            if( !_in_script_style  &&  _char == '>' )  
                            { 
                                _text_start = _entity_ptr,  _text_start.skip_char();
                                eat_char(); 
                                _in_script = false; 

                                if( _char == '\n' )         // Zeilenende hinter Skript ignorieren
                                {
                                    _text_start = _entity_ptr;
                                    _text_start.skip_char(); 
                                    eat_char();
                                }

                                goto SCRIPT_OK; 
                            }

                            _token._text += '%';


                default  :  _token._text += (char)eat_char();
            }
        }

SCRIPT_OK:;
        if( !_script_end )   _script_end = _entity_ptr;
    }
}

//--------------------------------------------------------Rtf_parser::convert_quotes
/*
void Rtf_parser::convert_quotes( string* str )
{
    int  n = str->length();

    for( int i = 0; i < n; i++ )
    {
        switch( (*str)[i] )
        {
            case '\x93': (*str)[i] = '"';  break;
            case '\x94': (*str)[i] = '"';  break;
            default: ;
        }
    }
}
*/

//---------------------------------------------------------------------synth_rtf_prop

static void synth_rtf_prop( void* synth, const Simple_entity& e )
{
    ((Synthesizer*)synth)->rtf_prop( (int)e.code(), e.param() );
}

//--------------------------------------------------------------Rtf_parser::put_props

void Rtf_parser::put_props( const Properties& props )
{
    if( (Simple_entity*)_text_start == _last_put_props )  return;

    props.write( _synth, synth_rtf_prop, (Flags)_allowed_properties ); 

    _last_put_props = _text_start;
}

//---------------------------------------------------------------Rtf_parser::has_text

bool Rtf_parser::has_text( Simple_entity* entity )
{
    if( !is_destination( entity->code() )   )  return false;
    if( Has_rtf_context::has_text( entity->code() ) )  return true;

    for( Entity_listelem* e = entity->head(); e; e = e->_next )
    {
        if( has_text(e) )  return true;
    }

    return false;
}

//--------------------------------------------------------------Rtf_parser::rtf_put_2

void Rtf_parser::rtf_put_2( Entity_listelem* start, Entity_listelem* end )
{
    // Gucken, ob Text in einer Destination (\header etc.) enthalten ist

    for( Entity_listelem* e = start; e != end; e = e->_next )
    {
        if( has_text(e) )
        {
            bool                is_script_synthesizer = _synth->is_script_synthesizer();
            bool                has_script;
            Synthesizer::Length synth_old_length = is_script_synthesizer? _synth->length() : Synthesizer::Length(0,0);

            _synth->rtf_put( (int)(Entity*)start, (int)(Entity*)e );
            _synth->rtf_open( (int)(Entity*)e );

            {
                Rtf_parser sub_parser ( has_rtf_context() );

                sub_parser._debug_outer_entity  = e;
                sub_parser._site                = _site;
                sub_parser._synth               = _synth;
                sub_parser._document_name       = "";
                sub_parser._include             = true;
                sub_parser._dont_merge          = true;   // RTF-Text hat keine \fonttbl etc., alle Nummern sind schon korrekt
                sub_parser._allowed_properties  = flag_prop;
                sub_parser._restore_properties  = false;
                sub_parser._template            = _template;
                sub_parser._template_index      = _template_index;
                sub_parser._entity_ptr          = e->head();
                sub_parser.set_constructor( _constructor );

                sub_parser.init();
                sub_parser.parse();

                has_script = sub_parser._has_script;        // Enthlt der RTF-Code Skript oder Ausdrcke?
                _has_script |= has_script;
            }

            _synth->rtf_close();

            if( is_script_synthesizer && has_script )  start = e->_next;
                                                 else  _synth->truncate( synth_old_length );     // Alles umsonst, der RTF-Text enthlt kein Skript.
        }
    }

    _synth->rtf_put( (int)(Entity*)start, (int)(Entity*)end );
}

//----------------------------------------------------------------Rtf_parser::rtf_put

void Rtf_parser::rtf_put( Entity_listelem* start, Entity_listelem* end )
{
    if( ( _allowed_properties & flag_prop ) == flag_prop )
    {
        rtf_put_2( start, end );
    }
    else  // Filtern
    {
        Entity_listelem* e = start;
        while( e != end )
        {
            start = e;

            while( e != end ) {
                if( descr( e->code() )->prop_flags() & ~_allowed_properties )  break;
                e = e->_next;
            }

            rtf_put_2( start, e );

            while( e != end ) {    // Nicht erlaubte Eigenschaften ignorieren:
                if( descr( e->code() )->prop_flags() & _allowed_properties )  break;
                e = e->_next;
            }
        }
    }
}

//------------------------------------------------------------------Rtf_parser::parse

void Rtf_parser::parse( const char* script_text )
{
    _template->read( script_text );
    _entity_ptr = _template->first_entity();
    parse();
}

//------------------------------------------------------------------Rtf_parser::parse

void Rtf_parser::parse()
{
    z::Log_ptr::Indent __indent__;
    if( z::Log_ptr log = "rtf.parse" )  *log << "Rtf_parser::parse {",  print_entity( log, *_debug_outer_entity  ),  *log << " ",  print_entity( log, *_entity_ptr ), *log << "..." << endl;

    Properties      saved_properties ( has_rtf_context() );
    Doc::Style*     style;

    _first_entity = _entity_ptr;

    // Der Dokumentkopf soll nicht im generierten Skript gelagert, sondern direkt erzeugt werden.
    // Damit ist _rtf_constructor->_doc->_header gefllt und auf die Verzeichnisse des Dokumentkopfs
    // kann zugegriffen werden.
    //jz 4.7.01 _skip_synth = !_include  &&  _rtf_constructor;    // _skip_synth = Dokument anfnglich direkt erzeugen, statt Umweg ber Skript zu nehmen

  //if( !_dont_merge  &&  _include  &&  _rtf_constructor )  merge();
    if( !_dont_merge  &&  _main_doc )  merge();

    // Dokumentkopf bernehmen (fr zusammengefasste Dokumente, 2.6.01)
    //jz 4.7.01 Entity_ptr first = _entity_ptr;
    //jz 4.7.01 while( _entity_ptr  &&  _entity_ptr->is_doc_property() )  _entity_ptr.to_next_entity();
    //jz 4.7.01 if( _allowed_properties & flag_prop_doc )  rtf_put( first, _entity_ptr );

    if( !_include )  _synth->rtf_start();

    if( _entity_ptr )
    {
        if( _include && _rtf_constructor )  
        {
            if( _restore_properties )  
            { 
                int f = _main_doc->_properties[code_f];
                saved_properties = _main_doc->_properties;
            }
        }

        _text_start = _entity_ptr._next_entity;

        eat_char();
        eat_token();

        // eat_char() stellt das erste Zeichen in _char bereit, also ist der Dokumentkopf verarbeitet.

        Doc::Header* header = _include? &_rtf_constructor->_doc->_header 
                                      : &_template->_header;

        style = header->style_by_name( "script" );
        if( style  &&  style->code() == code_s )  _script_para_style_no = style->_number;
                                            else  _script_para_style_no = -1;

        style = header->style_by_name( "script text" );
        if( style  &&  style->code() == code_cs )  _script_char_style_no = style->_number;
                                             else  _script_char_style_no = -1;



        // Dokumentkopf und anfngliche Eigenschaften ausgeben (zu Abschnittseigenschaften konvertierte Dokumenteigenschaften).
        // Ist nicht ntig, aber bersichtlicher. jz 27.10.04

        Simple_entity* before_body = _template->find_last_header_entity( _text_start );
        if( before_body != _text_start  &&  before_body  &&  before_body->_next )
        {
            // before_body darf nicht hinter _next_entity sein!
            // Das ist nicht mglich, weil _next_entity mit eat_char() Text gelesen hat und Text nicht zum Dokumentkopf gehrt.
            rtf_put( _text_start, before_body->_next );
            _text_start.forward_to( before_body->_next );
            put_props( _text_start._properties );
        }

        bool put_properties = false;


        while( _token._symbol != sym_eof )
        {
            // _text_start und _entity_ptr sollen auf eigene Entities zeigen (und nicht mittenrein)

            if( static_cast<Simple_entity*>( _entity_ptr ) == static_cast<Simple_entity*>( _text_start ) )      // Beide Zeiger zeigen in die selbe Entity
            {
                _entity_ptr._char_pos -= _text_start._char_pos;         // Bruchstelle
                _text_start.split_text();
                _entity_ptr._next_entity = _text_start._next_entity;    // _entity_ptr korrigieren
                _entity_ptr.split_text();
            } 
            else 
            {
                _text_start.split_text();
                _entity_ptr.split_text();
            }

            switch( _token._symbol )
            {
                case sym_text:          
                {
                    if( static_cast<Simple_entity*>( _text_start ) != static_cast<Simple_entity*>( _entity_ptr ) ) 
                    {
                        // Zuerst Stand der Eigenschaften ausgeben
                        if( put_properties )  put_props( _text_start._properties ), put_properties = false;

                        // Schlielich die als Zahl codierten Speicheradressen der Entities ausgeben
                        rtf_put( _text_start, _entity_ptr );
                    }

                    _text_start = _entity_ptr;
                    eat_token();
                    break;
                }


                case sym_script:
                    _has_script = true;
                                                                                                                                        
                    if( _rtf_constructor ) {
                        _script_end.correct_char_ptr_after_split();
                        _rtf_constructor->map_next_lines( _template_index, _synth->_next_lineno, _first_entity, _line_entity_ptr, _script_end, true );
                        _script_end = NULL;
                    }

                    _synth->put_script_line( _token._text );
                    eat_token();

                    put_properties = true;
                    break;


                case sym_expr:          
                    _has_script = true;

                    // Zuerst Stand der Eigenschaften ausgeben
                    if( put_properties )  put_props( _text_start._properties );

                    if( _rtf_constructor ) {
                        _script_end.correct_char_ptr_after_split();
                        _rtf_constructor->map_next_lines( _template_index, _synth->_next_lineno, _first_entity, _line_entity_ptr, _script_end, false );
                        _script_end = NULL;
                    }

                    _synth->put_expr( _token._text, _token._format );
                    eat_token();
                    break;


                case sym_preprocessor:  
                    _synth->put_preprocessor( _token._text );
                    eat_token();
                    put_properties = true;
                    break;


                default:                
                    throw_xc( "Rtf_parser::parse" );
            }
        }

        if( _include  &&  _rtf_constructor ) 
        {
            if( _restore_properties )  put_props( saved_properties );
        }
    }


    if( !_include )  _synth->rtf_finish();
}

//--------------------------------------------------------------------------------Rtf_parser::merge

void Rtf_parser::merge()
{
    //if( _allowed_properties & flag_prop_sect )  merge_properties( (rtf::Flags)_allowed_properties );   // \header bernehmen

    merge_table( code_stylesheet );
    merge_table( code_fonttbl );
    merge_table( code_listtable );
    merge_table( code_listoverridetable );
    merge_table( code_colortbl );

    modify_numbers( _template->first_entity() );
}

//----------------------------------------------------------------------Rtf_parser::map_merged_code

void Rtf_parser::map_merged_code( Code table_code, int number_in_template, int number_in_doc, const string& name )
{
    if( number_in_doc != number_in_template )
    {
        if( z::Log_ptr log = "rtf.merge" )  log << "\\" << descr(table_code)->_name << " " << number_in_template << " -> " << number_in_doc << " " << name << "\n";
    }

    number_map(table_code)[ number_in_template ] = number_in_doc;
}

//---------------------------------------------------------------------------Rtf_parser::add_to_doc

void Rtf_parser::add_to_doc( Code code, Entity_listelem* add, const string& debug_text )
{
    if( z::Log_ptr log = "rtf.merge" )  log << "add_to_doc " << *descr(code) << " ", print_entity( log, *add ), log << " " << debug_text << "\n";
    _main_doc->add_to_header_table( code, add, debug_text );
}

//--------------------------------------------------------------------------Rtf_parser::merge_table

void Rtf_parser::merge_table( Code table_code )
{
    if( table_code == code_listtable         )  { merge_listtable        ();  return; }
    if( table_code == code_listoverridetable )  { merge_listoverridetable();  return; }
    if( table_code == code_colortbl          )  { merge_colortbl         ();  return; }


    Doc::Header_map& template_map = _template->_header.header_map( table_code );
 //?Doc::Header_map& doc_map      = _main_doc->_header.header_map( table_code );

    Z_FOR_EACH( Doc::Header_map, template_map, template_it )
    if( template_it->second )
    {
        const Doc::Header_entry* template_entry = template_it->second;
        const Doc::Header_entry* doc_entry      = _main_doc->_header.entry_by_name( table_code, template_entry->_name );

        if( doc_entry )
        {
            map_merged_code( table_code, template_entry->_number, doc_entry->_number, template_entry->_name );
        }
        else
        {
            Destination_entity* entity = new Destination_entity( *cast_destination_entity(template_entry->_entity) );   // Kopie
            Entity_listelem*    e      = entity->head();

            if( table_code == code_fonttbl   ? e->code() == code_f_fonttbl :
                table_code == code_stylesheet? e->code() == code_s_stylesheet || e->code() == code_cs_stylesheet
                                             : false  ) 
            {
                e->set_param( _main_doc->_header.get_free_number( table_code, e->param() ) );
                map_merged_code( table_code, template_entry->_number, e->param(), template_entry->_name );
                add_to_doc( table_code, entity, template_entry->_name );

                _main_doc->build_table(table_code);
            }
            else
            {
                delete entity;
            }
        }
    }

    //_main_doc->build_table(table_code);
}

//-----------------------------------------------------------------Rtf_parser::change_id

void Rtf_parser::change_id( Simple_entity* se, map<int,bool>* used_map, map<int,int>* m, int* id )
{
    Param_entity* e = cast_param_entity( se );

    if( (*m)[ e->param() ] != 0 )        // Schonmal eine neue Id ausgedacht?
    {
        e->_param = (*m)[ e->param() ];
    }
    else
    {
        if( (*used_map)[ e->param() ] )  // Id schon im einbettenden Dokument verwendet?
        {
            while( (*used_map)[ *id ] )  ++*id;   // Neue Id ausdenken
            (*used_map)[ *id ] = true;

            (*m)[ e->param() ] = *id;
            e->_param = *id;
        }
        else
        {
            (*m)[ e->param() ] = e->param();
        }
    }
}

//----------------------------------------------------------------------Rtf_parser::merge_listtable

void Rtf_parser::merge_listtable()
{
    Doc::Header_map&  template_map = _template->_header.header_map( code_listtable );
    Doc::Header_map&  doc_map      = _main_doc->_header.header_map( code_listtable );
    
    int               listid          = 1000;
    int               listtemplateid  = 1000;
    int               leveltemplateid = 1000;


    // Feststellen, welche listtemplateid und leveltemplateid schon in Gebrauch sind:
    /* Nicht sperren, die Codes mssen erhalten bleiben! jz 4.7.01
    for( int i = -5; i <= 0; i++ ) {        // Einige Ids sind gesperrt
        _used_listid_map         [i] = true;
        _used_listtemplateid_map [i] = true;
        _used_leveltemplateid_map[i] = true;
    }
    */

    Z_FOR_EACH( Doc::Header_map, doc_map, it )
    if( it->second )
    {
        Doc::Header_entry*  entry = it->second;
        Entity_listelem*    e     = entry->_entity->head();

        _used_listtemplateid_map[ entry->_number ] = true;

        while(e)
        {
            if( e->code() == code_listtemplateid )  _used_listtemplateid_map[ e->param() ] = true;

            if( e->code() == code_listlevel ) 
            {
                Entity_listelem* f = e->head();
                while(f)
                {
                    if( f->code() == code_leveltext ) 
                    {
                        Entity_listelem* g = e->head();
                        while(g)
                        {
                            if( g->code() == code_leveltemplateid )  _used_leveltemplateid_map[ g->param() ] = true;
                            g = g->_next;
                        }
                    }
                    f = f->_next;
                }
            }

            e = e->_next;
        }
    }


    Z_FOR_EACH( Doc::Header_map, template_map, template_it )
    if( template_it->second )
    {
        Doc::Header_entry*  template_entry = template_it->second;
        Destination_entity* entity         = new Destination_entity( *cast_destination_entity(template_entry->_entity) );   // Kopie
        Entity_listelem*    e              = entity->head();
        int                 new_listid     = 0;

        while(e)
        {
            if( e->code() == code_listid )  change_id( e, &_used_listid_map, &_listid_map, &listid ), new_listid = e->param();
            else
            if( e->code() == code_listtemplateid )  change_id( e, &_used_listtemplateid_map, &_listtemplateid_map, &listtemplateid );
            else
            if( e->code() == code_listlevel ) 
            {
                Entity_listelem* f = e->head();
                while(f)
                {
                    if( f->code() == code_leveltext ) 
                    {
                        Entity_listelem* g = e->head();
                        while(g)
                        {
                            if( g->code() == code_leveltemplateid )  change_id( e, &_used_leveltemplateid_map, &_leveltemplateid_map, &leveltemplateid );
                            g = g->_next;
                        }
                    }
                    f = f->_next;
                }
            }

            e = e->_next;
        }

        add_to_doc( code_listtable, entity, as_string( new_listid ) );
        modify_numbers( entity, true );  // Z.B. \f in {\list...}, aber nicht \lists Verndern
    }

    _main_doc->build_table(code_listtable);
}

//--------------------------------------------------Rtf_parser::merge_listoverridetable

void Rtf_parser::merge_listoverridetable()
{
    Doc::Header_map& template_map = _template->_header.header_map( code_listoverridetable );

    Z_FOR_EACH( Doc::Header_map, template_map, template_it )
    if( template_it->second )
    {
        Doc::Header_entry*  template_entry = template_it->second;
        Destination_entity* entity         = new Destination_entity( *cast_destination_entity(template_entry->_entity) );   // Kopie
        Entity_listelem*    e              = entity->head();
        bool                have_ls        = false;
        int                 ls             = 0;

        while(e)
        {
            if( e->code() == code_ls ) 
            {
                if( !have_ls )  ls = _main_doc->_header.get_free_number( code_listoverridetable, e->param() ),  have_ls = true;
                e->set_param( ls );
                map_merged_code( code_listoverridetable, template_entry->_number, e->param() );
            }
            else
            if( e->code() == code_listid )  e->set_param( _listid_map[ e->param() ] );

            e = e->_next;
        }

        add_to_doc( code_listoverridetable, entity, as_string( ls ) );
        _main_doc->build_table(code_listoverride);
    }

    //_main_doc->build_table(code_listoverride);
}

//-----------------------------------------------------------------------Rtf_parser::merge_colortbl

void Rtf_parser::merge_colortbl()
{
    Doc::Header_map& template_map = _template->_header.header_map( code_colortbl );
    Doc::Header_map& doc_map      = _main_doc->_header.header_map( code_colortbl );

    Z_FOR_EACH( Doc::Header_map, template_map, template_it )
    if( template_it->second )
    {
        const Doc::Color*   template_color = (Doc::Color*)+template_it->second;
        const Doc::Color*   doc_color      = NULL;

        Z_FOR_EACH( Doc::Header_map, doc_map, doc_it )
        if( doc_it->second )
        {
            if( *(Doc::Color*)+doc_it->second == *(Doc::Color*)+template_it->second )
            {
                doc_color = (Doc::Color*)+doc_it->second;
                break;
            }
        }

        if( doc_color )
        {
          //number_map(code_colortbl)[ template_color->_number ] = doc_color->_number;
            map_merged_code( code_colortbl, template_color->_number, doc_color->_number );
        }
        else
        {
            int new_nr = _main_doc->_header.get_free_number( code_colortbl, template_color->_number );

            if( !template_color->_default )
            {
                add_to_doc( code_colortbl, new Param_entity( code_red  , template_color->_red   ) );
                add_to_doc( code_colortbl, new Param_entity( code_green, template_color->_green ) );
                add_to_doc( code_colortbl, new Param_entity( code_blue , template_color->_blue  ) );
            }

            add_to_doc( code_colortbl, new Text_entity( has_rtf_context(), ";" ) );

          //number_map(code_colortbl)[ template_color->_number ] = new_nr;
             map_merged_code( code_colortbl, template_color->_number, new_nr );

            _main_doc->build_table( code_colortbl );
        }
    }
}

//-----------------------------------------------------------------------Rtf_parser::modify_numbers

void Rtf_parser::modify_numbers( Entity_listelem* e, bool but_lists )
{
    while( e )
    {
        switch( e->code() )
        {
            case code_s              : 
            case code_snext          :
            case code_sbasedon       : 
            case code_cts            : 
            case code_cs             : 
            case code_ds             : e->set_param( number_map(code_stylesheet) [ e->param() ] );  break;

            case code_f              :
            case code_pnf            :
            case code_af             : e->set_param( number_map(code_fonttbl   ) [ e->param() ] );  break;

            case code_cf             :
            case code_pncf           : 
            case code_brdrcf         : 
            case code_cfpat          : 
            case code_cbpat          : 
            case code_clcfpat        : 
            case code_clcbpat        : 
            case code_cb             : 
            case code_ulc            : 
            case code_chcfpat        : 
            case code_chcbpat        : 
            case code_acf            : 
            case code_highlight      : e->set_param( number_map(code_colortbl  ) [ e->param() ] );  break;

            default:
                if( is_destination( e->code() ) )  modify_numbers( e->head(), but_lists );
                else
                if( !but_lists )
                switch( e->code() )
                {
                    case code_ls             : e->set_param( number_map(code_listoverridetable) [ e->param() ] );  break;

                    case code_listid         : e->set_param( _listid_map         [ e->param() ] );  break;
                    case code_listtemplateid : e->set_param( _listtemplateid_map [ e->param() ] );  break;
                    case code_leveltemplateid: e->set_param( _leveltemplateid_map[ e->param() ] );  break;
                    default:;
                }
        }

        e = e->_next;
    }
}

//-------------------------------------------------------------------------------------------------

} //namespace factory
} //namespace sos
