// $Id: factory_text.cxx 3546 2005-04-03 08:30:29Z jz $           Joacim Zschimmer

#include "precomp.h"

#include "../kram/sosstrng.h"
#include "../kram/sos.h"
#include "../kram/log.h"

#include "../zschimmer/z_com.h"
#include "../zschimmer/z_com_server.h"

#include "factory_processor.h"
#include "factory_text.h"


using namespace std;
using namespace zschimmer;
using namespace zschimmer::com;

namespace sos {
namespace factory {

//-------------------------------------------------------------------------------Typbibliothek

DESCRIBE_CLASS( &hostole_typelib, Text_constructor, factory_text_constructor , CLSID_Factory_text_constructor, 
                "DocumentFactoryC.Text_constructor", "1.0" );

//-----------------------------------------------------------------------Text_constructor::_methods
#ifndef SYSTEM_HAS_COM

const Com_method Text_constructor::_methods[] =
{ 
    { DISPATCH_METHOD, 1, "put"          , (Com_method_ptr)&Text_constructor::Put            , VT_EMPTY, { VT_VARIANT|VT_BYREF, VT_BOOL } },
    { DISPATCH_METHOD, 2, "put_formatted", (Com_method_ptr)&Text_constructor::Put_formatted  , VT_EMPTY, { VT_VARIANT|VT_BYREF, VT_INT } },
    { DISPATCH_METHOD, 3, "embed"        , (Com_method_ptr)&Text_constructor::Embed          , VT_EMPTY, { VT_BSTR } }, 
    { DISPATCH_METHOD, 4, "rtf_start"    , (Com_method_ptr)&Text_constructor::Rtf_start      , VT_EMPTY }, 
    { DISPATCH_METHOD, 5, "rtf_finish"   , (Com_method_ptr)&Text_constructor::Rtf_finish     , VT_EMPTY }, 
    { 0 }
};

#endif
//-----------------------------------------------------------Text_constructor::Text_constructor

Text_constructor::Text_constructor( Site* site ) 
: 
    Constructor( factory_text_constructor_class_ptr, this ),
    _zero_(this+1)
{
    _site = site;

  //_output_file.exceptions( ios::badbit | ios::failbit );
  //_output_file.open( _site->_output_filename, Any_file::out );
    _output_file.open( _site->_output_filename, O_WRONLY | O_CREAT | O_TRUNC, 0777 );
}

//----------------------------------------------------------Text_constructor::~Text_constructor

Text_constructor::~Text_constructor()
{
    LOGI( "~Text_constructor\n" );
}

//--------------------------------------------------------------------------Text_constructor::close

void Text_constructor::close()
{
    if( _output_file.opened() )
    {
        _output_file.close();
        _site->_document_written = true;
    }
}

//----------------------------------------------------------------Text_constructor::put_string

void Text_constructor::put_string( const string& text, bool raw )
{
    if( !raw  &&  _is_xml )
    {
        const char* p = text.c_str();

        while( *p )
        {
            const char* p0 = p;
            while( *p  &&  *p != '<'  &&  *p != '>'  &&  *p != '&' )  p++;
        
            _output_file.write( p0, p - p0 );

            if( !*p )  break;

            switch( *p ) 
            {
                case '<': _output_file.print( "&lt;"  );  p++;  break;
                case '>': _output_file.print( "&gt;"  );  p++;  break;
                case '&': _output_file.print( "&amp;" );  p++;  break;
              //case ' ': if( p == start  ||  p[-1] == ' ' )  { _buffer.append( "&#32;" ); p++; break; }  // Mehrere Leerzeichen
            }
        }
    }
    else
    {
        _output_file.print( text );
    }
}

//--------------------------------------------------------------------------Text_constructor::start

void Text_constructor::start()
{
}

//----------------------------------------------------------------------Text_constructor::rtf_start

STDMETHODIMP Text_constructor::Rtf_start()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        start();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Text_constructor::rtf_start" ); }

    return hr;
}

//-------------------------------------------------------------------------Text_constructor::finish

void Text_constructor::finish()
{
    close();
}

//---------------------------------------------------------------------Text_constructor::Rtf_finish

STDMETHODIMP Text_constructor::Rtf_finish()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        finish();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Text_constructor::rtf_finish" ); }

    return hr;
}

//-------------------------------------------------------------------------Text_parser::Text_parser

Text_parser::Text_parser()
: 
    _zero_(this+1)
{    
    obj_const_name( "factory::Text_parser" );
}
    
//------------------------------------------------------------------------Text_parser::~Text_parser

Text_parser::~Text_parser()
{    
    //if( _constructor )  _constructor->Release(),  _constructor = NULL;
}
    
//--------------------------------------------------------------------Text_parser::Text_parser

void Text_parser::init()
{    
    if( _constructor )
    {
        // Der Parser ist von #include oder embed() gerufen und soll den
        // bereits vorhandenen Konstruktor (Text_constructor oder Rtf_constructor) benutzen.
    }
    else
    {
        Constructor* c = new Text_constructor( _site );     if( !c )  throw_no_memory_error();
        set_constructor( c );
    }

    //?_constructor->AddRef();
}
    
//-----------------------------------------------------------------------Text_parser::eat_char

int Text_parser::eat_char()
{
    int result = _char;

    if( _char != EOF )
    {
        _eaten_char_pos._line = _char_pos._line;
        _eaten_char_pos._col  = _char_pos._col;
   
        if( _script_text_idx < length( _script_text ) ) 
        {
            _char = (unsigned char)_script_text[ _script_text_idx++ ];

            if( _char == '\n' )  _char_pos._line++,  _char_pos._col = 0;
                           else  _char_pos._col++;
        }
        else  
            _char = EOF;
    }

    return result;
}

//-----------------------------------------------------------------------Text_parser::eat_token

void Text_parser::eat_token()
{
    _token._pos    = _char_pos;
    _token._text   = "";
    _token._format = format_std;

    if( !_in_script )
    {
        if( _char == EOF )
        {
            _token._symbol = sym_eof;
        }
        else
        if( _char == _site->_var_char )
        {
            eat_char();

            if( _char == '$' )  _token._format = format_waehrung, eat_char();
        
            if( _char == '(' )
            {
                int klammern = 0;
            
                while(1)
                {
                    if( _char == EOF )  throw_xc( "EOF im &-Ausdruck" );
                    if( _char == '(' )  klammern++;
                    if( _char == ')' )  klammern--;

                    _token._text += (char)eat_char();

                    if( klammern == 0 )  break;
                }

                _token._symbol = sym_expr;
            }
            else
            {
                if( _site->_var_prefix )  _token._text = _site->_var_prefix;

                if( isalpha( _char ) )
                {
                    while( isalnum( (Byte)_char )  ||  _char == '_' )  _token._text += (char)eat_char();
                    _token._symbol = sym_expr;
                }
                else
                {
                    _token._text = _site->_var_char;
                    goto TEXT;
                }
            }
        }
        else
        if( _char == '<' )
        {
            eat_char();

            if( _char == '%' )
            {
                _in_script = true;
                eat_char();
                goto SCRIPT;
            }

            _token._text = "<";
            goto TEXT;
        }
        else
        {
TEXT:       
            _token._symbol = sym_text;

            while( _char != '<'  &&  _char != _site->_var_char  &&  _char != EOF )
            {
                _token._text += (char)eat_char();
            }
        }
    }
    else
    {
SCRIPT:
        _token._symbol = sym_script;
    
        while( isspace( _char ) )  eat_char();

        if( _char == '#' )  _token._symbol = sym_preprocessor,  eat_char();

        while(1)
        {
            switch( _char )
            {
                case EOF :  throw_xc( "EOF im Skript" );

                case '\r':  eat_char();  break;

                case '\n':  eat_char();  
                            goto SCRIPT_OK;

                case '%' :  eat_char();
                            
                            if( _char == '>' )  
                            { 
                                eat_char(); 
                                _in_script = false; 
                                if( _char == '\r' )  eat_char();    
                                if( _char == '\n' )  eat_char();    // Zeilenende gehrt zum Skript, nicht zum Text
                                goto SCRIPT_OK; 
                            }

                            _token._text += '%';

                default  :  _token._text += (char)eat_char();
            }
        }

SCRIPT_OK: ;
    }
}

//------------------------------------------------------------------Text_parser::parse

void Text_parser::parse( const char* script_text )
{   
    _script_text = script_text;

    eat_char();
    eat_token();

    if( _token._symbol == sym_text  
     &&  ( string_begins_with( _token._text, "<?xml"     ) 
        || string_begins_with( _token._text, "<!DOCTYPE" ) 
        || string_begins_with( _token._text, "<HTML"     ) 
        || string_begins_with( _token._text, "<html"     ) ) )
    {
        if( typeid( *_constructor ) == typeid (Text_constructor) )  ( static_cast<Text_constructor*>( +_constructor ) )->_is_xml = true;
    }

    while( _token._symbol != sym_eof )
    {
        switch( _token._symbol )
        {
            case sym_text:          _synth->put_text( _token._text, true );
                                    eat_token();
                                    break;


            case sym_script:        _synth->put_script_line( _token._text );
                                    eat_token();
                                    break;


            case sym_expr:          _synth->put_expr( _token._text, _token._format );
                                    eat_token();
                                    break;


            case sym_preprocessor:  _synth->put_preprocessor( _token._text );
                                    eat_token();
                                    break;


            default:                throw_xc( "Text_parser::parse" );
        }
    }

    if( !_include )  _synth->rtf_finish();
}

//-------------------------------------------------------------------------------------------------

} //namespace factory
} //namespace sos
