// $Id: factory_text.h 3546 2005-04-03 08:30:29Z jz $

#ifndef __FACTORY_TEXT_PARSER_H
#define __FACTORY_TEXT_PARSER_H

#include "../kram/sosobj.h"
//#include "../../file/flstream.h"
#include "../zschimmer/file.h"

namespace sos {
namespace factory {

//------------------------------------------------------------------------Text_constructor

struct Text_constructor : Ifactory_text_constructor,  // hostole.odl
                          Constructor
{
    void*                       operator new            ( size_t size )                     { return sos_alloc( size, "Text_constructor" ); }
    void                        operator delete         ( void* ptr )                       { sos_free( ptr ); }

                                Text_constructor        ( Site* );
                               ~Text_constructor        ();

    virtual void                close                   ();

    USE_SOS_OLE_OBJECT

    // IFactory_global_object:
    STDMETHODIMP                Put                     ( VARIANT* v, VARIANT_BOOL raw )    { return Constructor::Put( v, raw ); }
    STDMETHODIMP                Put_bstr                ( BSTR bstr, VARIANT_BOOL raw )     { return Constructor::Put_bstr( bstr, raw ); }
    STDMETHODIMP                Put_formatted           ( VARIANT* v, Ausgabeformat f )     { return Constructor::Put_formatted( v, f ); }
    STDMETHODIMP                Embed                   ( BSTR name )                       { return Constructor::Embed( name ); }
    STDMETHODIMP                Rtf_start               ();
    STDMETHODIMP                Rtf_finish              ();

    void                        put_string              ( const string&, bool raw );
    void                        start                   ();
    void                        finish                  ();

    Fill_zero                  _zero_;
    zschimmer::File            _output_file;            // Text-Ausgabe
    bool                       _is_xml;                 // "<" ">" "&" zu &lt; &gt; &amp;
};

//-----------------------------------------------------------------------------Text_parser

struct Text_parser : Parser
{
    BASE_CLASS( Parser )

    enum Symbol 
    {
        sym_none,
        sym_eof,
        sym_script,             // <% ... %>
        sym_preprocessor,       // <% #... %>
        sym_expr,               // &... oder &(...)
        sym_text,
    };

    struct Token
    {
        Symbol                 _symbol;
        Source_pos             _pos;
        Sos_string             _text;
        Ausgabeformat          _format;                 // Fr &$(...)
    };
    
                                Text_parser             ();
                               ~Text_parser             ();

    void                        init                    ();
    void                        parse                   ( const char* script_text );

    int                         eat_char                ();
    void                        eat_token               ();

    Bool                       _obj_is_type             ( Sos_type_code t ) const               { return t == tc_factory_Text_parser || Base_class::_obj_is_type( t ); }


    Fill_zero                  _zero_;

    Sos_string                 _script_text;
    int                        _script_text_idx;
    int                        _char;
    Source_pos                 _char_pos;
    Source_pos                 _eaten_char_pos;
    Token                      _token;
    bool                       _in_script;              // In <% ... %>

    Sos_string                 _saved_site_source;
    int                        _saved_site_first_lineno;
};


} //namespace factory
} //namespace sos


#endif
