// $Id: sos_mail_java.cxx 3952 2005-09-27 20:12:58Z jz $


#include "precomp.h"

#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>
#include <stdio.h>
#include <fcntl.h>

#ifdef SYSTEM_WIN
#   include <io.h>
#   include <share.h>
#endif

#include "sos.h"
#include "sosprof.h"
#include "sos_mail.h"
#include "sos_mail_jmail.h"
#include "com_simple_standards.h"
#include "sos_mail_java.h"
#include "../zschimmer/file.h"
#include "../zschimmer/java.h"


using zschimmer::vector_split;
using namespace zschimmer::java;

namespace sos {
namespace mail {

using namespace std;

//-----------------------------------------------------------------------Java_message::Java_message

Java_message::Java_message( Vm* vm )
:
    Global_jobject(vm), 
    _zero_(this+1),
    _env(vm),
    _java_class( vm, "sos/mail/Message" ),
    _close_method               ( &_java_class, "close"           , "()V" ),
    _set_method                 ( &_java_class, "set"             , "(Ljava/lang/String;[B)V" ),
    _get_method                 ( &_java_class, "get"             , "(Ljava/lang/String;)Ljava/lang/String;" ),
    _add_header_field_method    ( &_java_class, "add_header_field", "(Ljava/lang/String;Ljava/lang/String;)V" ),
    _add_attachment_method      ( &_java_class, "add_attachment"  , "([BLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V" ),
    _add_file_method            ( &_java_class, "add_file"        , "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V" ),
    _send_method                ( &_java_class, "send"            , "()V" )
{
    Env _env = env();
    
    jobject jo = _env->NewObject( _java_class, _java_class.method_id( "<init>", "()V" ) );
    if( !jo )  _env.throw_java( "NewObject", _java_class._name );

    assign( jo );

    if( _static->_debug )  set( "debug", "1" );

    Message::init();
}

//----------------------------------------------------------------------Java_message::~Java_message

Java_message::~Java_message()
{
    _env->CallVoidMethod( get_jobject(), _close_method );
    //if( _env->ExceptionCheck() )  _env.throw_java( "add_file" );
}

//-------------------------------------------------------------------------Java_message_object::set

void Java_message::set( const string& what, const char* value, int length )
{
    Local_frame local_frame ( jenv(), 10 );

    Jstring jwhat  ( vm(), what );
  //Jstring jvalue ( vm(), value, length );

    Local<jbyteArray> byte_jarray ( vm(), _env->NewByteArray( length ) );
    jbyte* byte_array = _env->GetByteArrayElements( byte_jarray, NULL );
    memcpy( byte_array, value, length );
    _env->ReleaseByteArrayElements( byte_jarray, byte_array, 0 );


    if( vm()->debug() )  LOG( "Java_message::set " << what << "=" << string(value,length) << "\n" );

    _env->CallVoidMethod( get_jobject(), _set_method._id, (jstring)jwhat, (jbyteArray)byte_jarray );
    if( _env->ExceptionCheck() )  _env.throw_java( "set", what + "=" + string(value,length) );
}

//-------------------------------------------------------------------------Java_message_object::get

string Java_message::get( const string& what, bool log_allowed )
{
    Local_frame local_frame ( jenv(), 10 );

    Jstring jwhat ( vm(), what );

    string result = _env.string_from_jstring( (jstring)_env->CallObjectMethod( get_jobject(), _get_method, (jstring)jwhat ) );

    if( log_allowed && vm()->debug() )  LOG( "Java_message::get " << what << "==>" << result << '\n' );
    if( _env->ExceptionCheck() )  _env.throw_java( "get", what );

    return result;
}

//---------------------------------------------------------------------------Java_message::set_from

void Java_message::set_from( const string& from )
{
    set( "from", from );
}

//-------------------------------------------------------------------------------Java_message::from

string Java_message::from()
{
    return get( "from" );
}

//-----------------------------------------------------------------------Java_message::set_reply_to

void Java_message::set_reply_to( const string& reply_to )
{
    set( "reply-to", reply_to );
}

//---------------------------------------------------------------------------Java_message::reply_to

string Java_message::reply_to()
{
    return get( "reply-to" );
}

//-----------------------------------------------------------------------Java_message::set_subject_

void Java_message::set_subject_( const string& subject )
{
    set( "subject", subject.substr( 0, 200 ) );
}

//----------------------------------------------------------------------------Java_message::subject

string Java_message::subject()
{
    return get( "subject" );
}

//---------------------------------------------------------------------------Java_message::set_body

void Java_message::set_body( const string& body )
{
    set( "body", body );
}

//-------------------------------------------------------------------------------Java_message::body

string Java_message::body()
{
    return get( "body" );
}

//-------------------------------------------------------------------Java_message::set_content_type

void Java_message::set_content_type( const string& content_type )
{
    //if( content_type != ""  &&  content_type != "text/plain" )  throw_xc( "Java_message::set_content_type" );
    set( "content_type", content_type );
}

//-----------------------------------------------------------------------Java_message::set_encoding

void Java_message::set_encoding( const string& encoding )
{
    //throw_xc( "Java_message::set_encoding" );
    //LOG( "Java_message::set_encoding(\"" << encoding << "\") wird fr JavaMail ignoriert\n" );
    set( "encoding", encoding );
}

//--------------------------------------------------------------------------Java_message::add_file_

void Java_message::add_file_( const string& real_filename, const string& new_filename, const string& content_type, const string& encoding )
{
    Local_frame local_frame ( jenv(), 10 );

    Jstring real_filename_j ( vm(), real_filename     );
    Jstring new_filename_j  ( vm(), new_filename     );
    Jstring content_type_j  ( vm(), content_type.c_str() );     //content_type.empty()? "text/plain" : 
    Jstring encoding_j      ( vm(), encoding     );

    if( vm()->debug() )  LOG( "Java_message::add_file()\n" );

    _env->CallVoidMethod( get_jobject(), _add_file_method, 
                          (jstring)real_filename_j, (jstring)new_filename_j, (jstring)content_type_j, (jstring)encoding_j );
    if( _env->ExceptionCheck() )  _env.throw_java( "add_file" );

/*
    try
    {
        zschimmer::Mapped_file file ( real_filename, "r" );

        add_attachment( file.map(), file.length(), 
                        mail_filename != ""? mail_filename : filename_of_path( real_filename ), 
                        content_type, encoding );
    }
    catch( const exception& x )  { throw_xc( x ); }
*/
}

//---------------------------------------------------------------------Java_message::add_attachment

void Java_message::add_attachment( const string& data, const string& filename, const string& content_type, const string& encoding )
{
    add_attachment( data.data(), data.length(), filename, content_type, encoding );
}

//---------------------------------------------------------------------Java_message::add_attachment

void Java_message::add_attachment( const void* data, int length, const string& filename, const string& content_type, const string& encoding )
{
    Local_frame local_frame ( jenv(), 10 );

    Local<jbyteArray> byte_jarray ( vm(), _env->NewByteArray( length ) );
    jbyte* byte_array = _env->GetByteArrayElements( byte_jarray, NULL );
    memcpy( byte_array, data, length );
    _env->ReleaseByteArrayElements( byte_jarray, byte_array, 0 );

    Jstring data_j         ( vm(), (const char*)data, length );
    Jstring filename_j     ( vm(), filename     );
    Jstring content_type_j ( vm(), content_type.c_str() );    //content_type.empty()? "text/plain" : 
    Jstring encoding_j     ( vm(), encoding     );

    if( vm()->debug() )  LOG( "Java_message::add_attachment()\n" );

    _env->CallVoidMethod( get_jobject(), _add_attachment_method, 
                          (jbyteArray)byte_jarray, 
                          (jstring)filename_j, (jstring)content_type_j, (jstring)encoding_j );
    if( _env->ExceptionCheck() )  _env.throw_java( "add_attachment" );
}

//-------------------------------------------------------------------Java_message::set_header_field

void Java_message::add_header_field( const string& field_name, const string& value )
{
    Local_frame local_frame ( jenv(), 10 );

    Jstring field_name_j ( vm(), field_name );
    Jstring value_j      ( vm(), value      );

    if( vm()->debug() )  LOG( "Java_message::add_header_field " << field_name << ": " << value << '\n' );

    _env->CallVoidMethod( get_jobject(), _add_header_field_method, (jstring)field_name_j, (jstring)value_j );
    if( _env->ExceptionCheck() )  _env.throw_java( "add_header_field" );
}

//------------------------------------------------------------------------------Java_message::send2

void Java_message::send2()
{
    LOG( "JavaMail Send smtp=" << get("smtp",false) << " to=\"" << get("to",false) << "\" subject=\"" << get("subject",false) << "\"\n" );

    _env->CallVoidMethod( get_jobject(), _send_method.id() );
    if( _env->ExceptionCheck() )  _env.throw_java( "send" );

    LOG( "JavaMail Send fertig\n" );
}

//------------------------------------------------------------------------Java_message::rfc822_text

string Java_message::rfc822_text()
{
    return get( "rfc822_text" );
}

//------------------------------------------------------------------------Java_message::send_rfc822

void Java_message::send_rfc822( const char* rfc822_text, int length )
{
    set( "send_rfc822", rfc822_text, length );
}

//-------------------------------------------------------------------------------------------------

} //namespace mail
} //namespace sos
