// $Id: async_socket.h 4057 2006-01-25 08:19:41Z jz $

#ifndef __Z_ASYNC_SOCKET_H
#define __Z_ASYNC_SOCKET_H

#include "async.h"
#include "z_sockets.h"

//-------------------------------------------------------------------------------------------------

namespace zschimmer { 

struct Socket_manager;

//--------------------------------------------------------------------------------------Socket_wait
#ifndef Z_WINDOWS

struct Socket_wait : Object
{
    enum Read_or_write
    { 
        readfd, 
        writefd,
        exceptfd
    };

                                Socket_wait                 ( Socket_manager* );
                              //Socket_wait                 ( Socket_manager*, int n, const fd_set& readfds, const fd_set& writefds, const fd_set& exceptfds );

  //void                        add                         ( int file_nr, Read_or_write, Has_set_signaled* );
    void                        add                         ( Event_base* );
    int                         wait                        ( double seconds );
    void                    set_polling_interval            ( double seconds )                      { _polling_interval = seconds; }

    Fill_zero                  _zero_;
    Socket_manager*            _socket_manager;
  //int                        _n;
  //fd_set                     _fds[3];
  //std::vector<Has_set_signaled*>  _signaled_array;

    typedef std::list<Event_base*>   Event_list;
    Event_list                 _event_list;
    double                     _polling_interval;
};

#endif

//-------------------------------------------------------------------------------------Socket_event

#ifdef Z_WINDOWS
    typedef Event               Socket_event;
# else
    typedef Simple_event        Socket_event;
#endif

//---------------------------------------------------------------------------------Socket_operation

struct Socket_operation : Async_operation
{
    enum Signals
    {
        sig_none   = 0,
        sig_read   = 0x01,
        sig_write  = 0x02,
        sig_except = 0x04
    };


                                Socket_operation            ( Socket_manager* = NULL );
                               ~Socket_operation            ();

    virtual void                close                       ();

    void                        add_to_socket_manager       ( Socket_manager* );
    void                        remove_from_socket_manager  ();
    void                    set_blocking                    ( bool b )                              { _blocking = b; }
  //void                        listen                      ();
    bool                        accept                      ( SOCKET listen_socket );

    void                        call_ioctl                  ( int what, unsigned long value );
    void                    set_keepalive                   ( bool );
    void                    set_linger                      ( bool, int seconds = 0 );

    Socket_event*               async_event                 ()                                      { return &_socket_event; }
    void                    set_event_name                  ( const string& name )                  { _socket_event.set_name( name ); }

    void                        socket_expect_signals       ( Signals );

  //void                    set_socket_signals              ( Signals signals )                     { _socket_signals = signals; }
  //void                  clear_socket_signals              ()                                      { _socket_signals = sig_none; }

    bool                        async_signaled_             ();
    void                        async_clear_signaled_       ();
/*
    bool                        socket_read_signaled        ()                                      { return ( _socket_signals & sig_read   ) != 0; }
    bool                        socket_write_signaled       ()                                      { return ( _socket_signals & sig_write  ) != 0; }
    bool                        socket_except_signaled      ()                                      { return ( _socket_signals & sig_except ) != 0; }
*/
    string                      obj_name                    () const                                { return "TCP " + _peer_host_and_port; }
    const Host_and_port         peer                        () const                                { return _peer_host_and_port; }

//protected:
 // friend struct Socket_manager;
    void                        set_socket_event_name       ();

    Fill_zero _zero_;

    SOCKET                     _read_socket;
    SOCKET                     _write_socket;               // _read_socket != _write_socket, nur wenn _read_socket==stdin und _write_socket==stdout
    bool                       _eof;                        // recv(_read_socket) hat 0 geliefert, also die Verbindung geschlossen
  //Signals                    _socket_signals;
    Socket_manager*            _socket_manager;
    Socket_event               _socket_event;               // In Unix nur bei Http_response/Log_chunk_reader

  //struct sockaddr_in         _peer_addr;
    Host_and_port              _peer_host_and_port;
    bool                       _blocking;

#   ifdef Z_WINDOWS
        int                    _wsa_event_select_flags;
#   endif
};

inline Socket_operation::Signals  operator |  ( Socket_operation::Signals  a, Socket_operation::Signals b )  { return (Socket_operation::Signals)( (int)a | (int)b ); }
inline Socket_operation::Signals& operator |= ( Socket_operation::Signals& a, Socket_operation::Signals b )  { a = a | b; return a; }

//------------------------------------------------------------------------Buffered_socket_operation

struct Buffered_socket_operation : Socket_operation
{
    enum State 
    { 
        s_initial, 
        s_connecting,
        s_stand_by,             
        s_sending, 
        s_waiting,              // recv(), noch keine Daten empfangen
        s_receiving,            // recv(), bereits Daten empfangen
        s_processing,           // Empfangene Daten werden von oberer Klasse verarbeitet
      //s_closed,
    };


                                Buffered_socket_operation   ( Socket_manager* = NULL );

    void                        close                       ();

    void                        connect__start              ( const Host_and_port& );
    bool                        connect__continue           ();

    void                        set_buffer_size             ();
    bool                        send__start                 ( const string& data );
    bool                        send__continue              ();

    int                         call_recv                   ( char* buffer, int size );
    bool                        recv__start                 ();
    bool                        recv__continue              ();
    const string&               recv_data                   ()                                      { return _recv_data; }
    void                        recv_clear                  ()                                      { _recv_data = ""; }
    void                        assert_no_recv_data         ();

    void                    set_state                       ( State state )                         { _state = state; }
    State                       state                       () const                                { return _state; }
    string                      state_name                  () const                                { return state_name( _state ); }
    static string               state_name                  ( State );

  protected:
    bool                        async_continue_             ( Continue_flags );
    bool                        async_has_error_            ()                                      { return false; }
    bool                        async_finished_             ()                                      { return _state == s_initial  
                                                                                                          || _state == s_stand_by; }
    string                      async_state_text_           ();


  private:
    Fill_zero                  _zero_;
    State                      _state;
    int                        _socket_send_buffer_size;
    string                     _send_data;
    size_t                     _send_progress;
    string                     _recv_data;
};

//-----------------------------------------------------------------------------------Socket_manager

struct Socket_manager : Async_manager
{
    enum Read_or_write
    { 
        read_fd, 
        write_fd,
        except_fd
    };


                                Socket_manager              ();
                               ~Socket_manager              ();


    void                        get_events                  ( std::vector<Socket_event*>* );

#ifdef Z_WINDOWS
    void                        set_fd                      ( Read_or_write, SOCKET )               {}
    void                        clear_fd                    ( Read_or_write, SOCKET )               {}
    void                        clear_socket_signaled       ( SOCKET )                              {}
#else
    void                        set_fd                      ( Read_or_write rw, SOCKET s )          { FD_SET( s, &_fds[rw] );  if( _n < s+1 )  _n = s+1; }
    void                        clear_fd                    ( Read_or_write rw, SOCKET s )          { FD_CLR( s, &_fds[rw] ); }
    ptr<Socket_wait>            create_wait                 ();
    bool                        socket_signaled             ( SOCKET s )                            { return FD_ISSET( s, &_signaled_fds[read_fd  ] ) ||
                                                                                                             FD_ISSET( s, &_signaled_fds[write_fd ] ) ||
                                                                                                             FD_ISSET( s, &_signaled_fds[except_fd] );   }
    void                        clear_socket_signaled       ( SOCKET s )                            { FD_CLR( s, &_signaled_fds[read_fd] );
                                                                                                      FD_CLR( s, &_signaled_fds[write_fd] );
                                                                                                      FD_CLR( s, &_signaled_fds[except_fd] ); };
#endif

    int                         wait                        ( double timeout_seconds );

    void                        set_read_fd                 ( SOCKET s )                            { set_fd( read_fd, s ); }
    void                        set_write_fd                ( SOCKET s )                            { set_fd( write_fd, s ); }
    void                        set_except_fd               ( SOCKET s )                            { set_fd( except_fd, s ); }

    bool                        async_continue              ()                                      { return async_continue_selected( NULL ); }
    bool                        async_continue_selected     ( Operation_is_ok f, double sleep = 0.0 );


  protected:
    friend struct               Socket_operation;

    // Benutze Socket_operation::add_to_socket_manager()!
    void                        add_socket_operation        ( Socket_operation* );
    void                        remove_socket_operation     ( Socket_operation* );

    bool                        async_continue_selected_    ( Operation_is_ok* );


    Fill_zero                  _zero_;

    typedef std::list< Socket_operation* >  Socket_operation_list;
  //typedef std::map< SOCKET, Socket_operation* >  Socket_operation_map;
    Socket_operation_list      _socket_operation_list;
    bool                       _socket_operation_list_modified;


#   ifndef Z_WINDOWS
        SOCKET                 _n;
        fd_set                 _fds[3];
        
        SOCKET                 _socket_signaled_count;
        fd_set                 _signaled_fds[3];
#   else
        bool                   _wsastartup_called;
#   endif
};

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

#endif
