// $Id: java.h 4038 2005-12-15 10:02:25Z jz $

#ifndef __ZSCHIMMER_JAVA_H_
#define __ZSCHIMMER_JAVA_H_


#include <vector>
#include <jni.h>
#include "threads.h"
#include "z_com.h"
#include "log.h"


namespace zschimmer {
namespace java {

const int                       max_params                  = 10;

struct Vm;
struct Method;

//-------------------------------------------------------------------------------------------------

string                          string_from_jstring         ( JNIEnv*, const jstring& );
jstring                         jstring_from_string         ( JNIEnv*, const char* str, size_t length );
inline jstring                  jstring_from_string         ( JNIEnv* jenv, const char* str )       { return jstring_from_string( jenv, str, strlen(str) ); }
inline jstring                  jstring_from_string         ( JNIEnv* jenv, const string& str )     { return jstring_from_string( jenv, str.data(), str.length() ); }
Vm*                             create_vm                   ( bool init = true );
Z_NORETURN void                 throw_java_ret              ( int return_value, const string& method_name, const string& = "" );
void                            set_java_exception          ( JNIEnv*, const char* what );
void                            set_java_exception          ( JNIEnv*, const exception& );
void                            set_java_exception          ( JNIEnv*, const _com_error& );
void                            set_NullPointerException    ( JNIEnv*, const string& text = "" );

//-----------------------------------------------------------------------------------Java_exception

struct Java_exception : Xc
{
                                Java_exception              ( const char* code )                    : Xc(code) {}
    Z_GNU_ONLY(                ~Java_exception              () throw()                              {}  )               // gcc 3.2.2

    string                      name                        () const                                { return _exception_name; }
    string                      message                     () const                                { return _message; }

    string                     _exception_name;
    string                     _message;
};

//----------------------------------------------------------------------------------------------Env

struct Env 
{
                                Env                         ( JNIEnv* jenv )                        : _jenv( jenv ) {}
                                Env                         ( Vm* );

                                operator JNIEnv*            ()                                      { return jenv(); }
                                operator JavaVM*            ()                                      { return vm(); }
    JNIEnv*                     operator ->                 ()                                      { return jenv(); }

    JNIEnv*                     jenv                        ()                                      { return _jenv; }
    JavaVM*                     vm                          ()                                      { JavaVM* vm = NULL; _jenv->GetJavaVM(&vm); return vm; }

    Z_NORETURN void             throw_java                  ( const string&, const string& = "" );
  //Z_NORETURN void             throw_java                  ( jclass, const string& = "" );
    
  //void                        add_object                  ( Java_idispatch* );
  //void                        release_objects             (); //?                                      { _java_idispatch_list.clear(); }

    // Einige Java-Methoden mit Fehlerprfung:
    jclass                      find_class                  ( const string& name );
    jmethodID                   get_method_id               ( jclass cls, const string& name, const string& signature, bool is_static = false ) { return get_method_id( cls, name.c_str(), signature.c_str(), is_static ); }
    jmethodID                   get_method_id               ( jclass, const char* name, const char* signature, bool is_static = false );
    jclass                      get_object_class            ( jobject );
    string                      string_from_jstring         ( jstring jstr )                        { return ::zschimmer::java::string_from_jstring( _jenv, jstr ); }


    JNIEnv*                    _jenv;
    //?std::list< ptr<Java_idispatch> >  _java_idispatch_list;     // Hlt alle in einer native Methode erzeugten IDispatchs, bis release_objects()
    //std::list< Java_idispatch* >  _java_idispatch_list;     // Hlt alle in einer native Methode erzeugten IDispatchs, bis release_objects()
};

//-----------------------------------------------------------------------------------------------Vm
// Fr dieselbe JavaVM knnen mehrere VM vorhanden sein, fr jede DLL eine.
// Neben dem Hauptprogram zurzeit hostole.dll, spidermonkey.dll, hostjava.dll.

struct Vm : Object              // Java virtual machine
{
    struct Option
    {
                                Option                      ( const string& option, void* extra = NULL ) : _option(option), _extra(extra) {}

        string                 _option;
        void*                  _extra;
    };


    struct Instance
    {
                                Instance                    ( const string& class_name, jobject jo ) : _class_name(class_name), _jobject(jo) {}

        string                 _class_name;
        jobject                _jobject;
    };



                                Vm                          ( bool do_start = true );
                                Vm                          ( JavaVM* );                            // Vorhandene JVM verwenden (fr natives Module)
                               ~Vm                          ();


    void                    set_debug                       ( bool d )                              { _debug = d; }
    bool                        debug                       ()                                      { return _debug; }

    bool                        running                     () const                                { return _vm != NULL; }

    void                        new_instances               ( const string& class_names );          // "class:class:..."
    void                    set_log                         ( Has_log* log )                        { _log.set_log( log ); }

    void                    set_filename                    ();
    void                    set_filename                    ( const string& module_filename )       { _filename = module_filename; }
    string                      filename                    () const                                { return _filename; }

    void                    set_class_path                  ( const string& class_path )            { _class_path = class_path; }
    string                      class_path                  ()                                      { return _class_path; }
    void                        append_class_path           ( const string& path )                  { _class_path += Z_PATH_SEPARATOR; _class_path += path; }
    void                        prepend_class_path          ( const string& path )                  { _class_path = path + Z_PATH_SEPARATOR + _class_path; }

    void                    set_javac_filename              ( const string& filename )              { _javac_filename = filename; }
    string                      javac_filename              () const                                { return _javac_filename; }

    void                    set_work_dir                    ( const string& directory )             { _work_dir = directory; }
    string                      work_dir                    ()                                      { return _work_dir; }

    void                    set_destroy_vm                  ( bool );

    void                    set_sos_initialized             ( bool b = true )                       { _sos_initialized = b; }       // Fr kram/sos_java.h
    bool                        sos_initialized             () const                                { return _sos_initialized; }

    void                        init                        ();
    void                        start                       ();
    void                        close                       ();
    void                        log_version                 ( JNIEnv*, const string& module_filename );
    void                    set_options                     ( const string& );

    JNIEnv*                     attach_thread               ( const string& thread_name );
    void                        detach_thread               ();

    JavaVM*                     vm                          ()                                      { return _vm; }

    Env                         env                         ()                                      { return Env( jenv() ); }
    JNIEnv*                     jenv                        ();

    void                        check_for_exception         ();

  //void                        add_object                  ( Java_idispatch* );
  //void                        release_objects             ()                                      { _java_idispatch_list.clear(); }


  //static void             set_jvm                         ( JavaVM* );
  //static JavaVM*              jvm                         ();
    static ptr<Vm>          get_vm                          ( bool start = true );
    static JavaVM*              request_jvm                 ();
    static void                 release_jvm                 ( JavaVM* );

    static Mutex                static_vm_mutex;
    static Vm*                  static_vm;
  //static JavaVM*              static_foreign_jvm;
  //static JavaVM*              static_jvm;                 // In java_jvm.cxx


    Fill_zero                  _zero_;
    bool                       _foreign;                    // JVM ist von auen (fr Hostjava)
    bool                       _vm_requested;
    bool                       _dont_destroy;               // DestroyJavaVM() nicht rufen (hngt manchmal)
    bool                       _debug;
    bool                       _sos_initialized;
    Delegated_log              _log;
    JavaVMInitArgs             _vm_args;
    std::vector<Option>        _options;
    JavaVM*                    _vm;
  //Thread_data<Java_thread_data> _thread_data;

    string                     _filename;                   // Dateiname der Java-VM
    string                     _class_path;
    string                     _complete_class_path;
    string                     _javac_filename;
    string                     _work_dir;
    string                     _new_instances;

    typedef std::vector<Instance> Instances;
    Instances                  _instances;

  //jclass                     _idispatch_jclass;
  //list< ptr<Java_idispatch> >_java_idispatch_list;        // Hlt alle in einer nativen Methode erzeugten IDispatchs, bis release_objects()
};

//-------------------------------------------------------------------------------------------Has_vm
// inlines am Ende des Headers

struct Has_vm
{
                                Has_vm                      ( Vm* );

    JNIEnv*                     jenv                        ();
    Env                         env                         ();
    Vm*                         vm                          ();

    void                    set_vm                          ( Vm* );

    ptr<Vm>                    _java_vm;
};

//--------------------------------------------------------------------------------------Local_frame

struct Local_frame
{
                                Local_frame                 ( JNIEnv*, int capacity = 20 );
                               ~Local_frame                 ()                                      { _jenv->PopLocalFrame(NULL); }

    JNIEnv*                    _jenv;
};

//--------------------------------------------------------------------------------------Jvalue_kind
/*
struct Jvalue_kind
{
    jvalue_none = 0,
    jvalue_z    = 'Z',
    jvalue_b    = 'B',
    jvalue_c    = 'C',
    jvalue_s    = 'S',
    jvalue_i    = 'I',
    jvalue_j    = 'J',
    jvalue_f    = 'F',
    jvalue_d    = 'D',
    jvalue_l
};
*/
//-------------------------------------------------------------------------------------------Jvalue

struct Jvalue
{
                                Jvalue                      ( jobject value )                       : _type( 'L' ) { _jvalue.l = value; }
                                Jvalue                      ( jboolean value )                      : _type( 'Z' ) { _jvalue.z = value; }
                                Jvalue                      ( bool value )                          : _type( 'Z' ) { _jvalue.z = value; }
                                Jvalue                      ( char value )                          : _type( 'C' ) { _jvalue.c = value; }
                                Jvalue                      ( jshort value )                        : _type( 'S' ) { _jvalue.s = value; }
                                Jvalue                      ( jint value )                          : _type( 'I' ) { _jvalue.i = value; }
                                Jvalue                      ( jlong value )                         : _type( 'j' ) { _jvalue.j = value; }
                                Jvalue                      ( jfloat value )                        : _type( 'F' ) { _jvalue.f = value; }
                                Jvalue                      ( jdouble value )                       : _type( 'D' ) { _jvalue.d = value; }

    const jvalue&               get_jvalue                  ()                                      { return _jvalue; }

    jobject                     as_jobject                  ()                                      { check( 'L' );  return _jvalue.l; }
    bool                        as_bool                     ()                                      { check( 'Z' );  return _jvalue.b != 0; }

    void                        check                       ( char type );

    jvalue                     _jvalue;
    char                       _type;
};

//------------------------------------------------------------------------------------------Jobject

struct Jobject : Has_vm, Jvalue, Non_cloneable
{
    enum Registered { reg_none, reg_local, reg_global };

    Z_GNU_ONLY( private:        Jobject                     (); public: )                           // Fr gcc 3.2. Nicht implementiert.
                                Jobject                     ( Vm* vm, jobject = NULL );
    virtual                    ~Jobject                     ();                                     // Fr gcc 3.2

  //STDMETHODIMP                QueryInterface              ( const IID&, void** );

                                operator jobject            ()                                      { return get_jobject(); }
    Jobject&                    operator =                  ( jobject jo )                          { assign( jo );  return *this; }

  //void                        set_local                   ();
    void                        set_global                  ();

    virtual jclass              get_jclass                  ();

    virtual void                assign                      ( jobject );

    jobject                     get_jobject                 ()                                      { return _jvalue.l; }

    jmethodID                   method_id                   ( const char*   name, const char*   signature );
    jmethodID                   method_id                   ( const string& name, const string& signature ) { return method_id( name.c_str(), signature ); }

    jmethodID                   static_method_id            ( const char*   name, const char*   signature );
    jmethodID                   static_method_id            ( const string& name, const string& signature ) { return static_method_id( name.c_str(), signature.c_str() ); }

    void                        call_void_method            ( const char* name, const char* signature, ... );
    jobject                     call_object_method          ( const char* name, const char* signature, ... );
    jint                        call_int_method             ( const char* name, const char* signature, ... );
    jlong                       call_long_method            ( const char* name, const char* signature, ... );
    bool                        call_bool_method            ( const char* name, const char* signature, ... );
    bool                        call_bool_method            ( jmethodID, ... );
    string                      call_string_method          ( const char* name, const char* signature, ... );
    string                      call_string_method          ( jmethodID, ... );

    jobject                     call_static_object_method   ( const char* name, const char* signature, ... );

/*
    Jvalue                      call                        ( const char* name_and_signature );
    Jvalue                      call                        ( const char* name_and_signature, const Jvalue& );
    Jvalue                      call                        ( const char* name_and_signature, const Jvalue&, const Jvalue& );
    Jvalue                      call                        ( const char* name_and_signature, const Jvalue&, const Jvalue&, const Jvalue& );
*/
  //jobject                    _jobject;
    Registered                 _registered;
};

//-----------------------------------------------------------------------------------Global_jobject

struct Global_jobject : Jobject
{
                                Global_jobject              ( Vm* vm, jobject jo = NULL )           : Jobject( vm, NULL )  { assign( jo ); }
                               ~Global_jobject              ()                                      { assign( NULL ); }

    Global_jobject&             operator =                  ( jobject jo )                          { assign( jo );  return *this; }

    void                        assign                      ( jobject );
};

//-------------------------------------------------------------------------------------------Global

template< class CLASS >
struct Global : Global_jobject
{
                                Global                      ( Vm* vm, CLASS jo = NULL )             : Global_jobject( vm, NULL )  { assign( jo ); }

                                operator CLASS              ()                                      { return (CLASS)get_jobject(); }
    CLASS&                      operator =                  ( CLASS jo )                            { assign( jo );  return *this; }
    void                        assign                      ( CLASS jo )                            { Global_jobject::assign( jo ); }
};

//------------------------------------------------------------------------------------Local_jobject

struct Local_jobject : Jobject
{
                                Local_jobject               ( Vm* vm, jobject jo = NULL )           : Jobject( vm, NULL )  { assign( jo ); }
                               ~Local_jobject               ()                                      { assign( NULL ); }

    Local_jobject&              operator =                  ( jobject jo )                          { assign( jo );  return *this; }

    void                        assign                      ( jobject );
};

//--------------------------------------------------------------------------------------------Local

template< class CLASS >
struct Local : Local_jobject
{
                                Local                       ( Vm* vm, CLASS jo = NULL )             : Local_jobject( vm, NULL )  { assign( jo ); }

                                operator CLASS              ()                                      { return (CLASS)get_jobject(); }
    CLASS&                      operator =                  ( CLASS jo )                            { assign( jo );  return *this; }
    void                        assign                      ( CLASS jo )                            { Local_jobject::assign( jo ); }
};

//------------------------------------------------------------------------------------------Jstring

struct Jstring : Has_vm, Jvalue, Non_cloneable
{
    Z_GNU_ONLY(                 Jstring                     ();  )                                  // Fr gcc 3.2. Nicht implementiert.
                                Jstring                     ( Vm* vm )                              : Has_vm(vm), Jvalue( (jstring)NULL ) {}
                                Jstring                     ( Vm* vm, jstring jstr )                : Has_vm(vm), Jvalue( (jstring)NULL ) { assign( jstr ); }
                                Jstring                     ( Vm* vm, const string& str )           : Has_vm(vm), Jvalue( (jstring)NULL ) { assign( str ); }
                                Jstring                     ( Vm* vm, const char* str )             : Has_vm(vm), Jvalue( (jstring)NULL ) { assign( str ); }
                                Jstring                     ( Vm* vm, const char* str, size_t length ) : Has_vm(vm), Jvalue( (jstring)NULL ) { assign( str, length ); }
                               ~Jstring                     ();

  //STDMETHODIMP                QueryInterface              ( const IID&, void** );

    void                        operator =                  ( jstring jstr )                        { assign( jstr ); }
    void                        operator =                  ( const string& str )                   { assign( str ); }
    void                        operator =                  ( const char* str )                     { assign( str ); }
                                operator jstring            ()                                      { return get_jstring(); }

    void                        assign                      ( jstring );
    void                        assign                      ( const string& str )                   { assign( str.data(), str.length() ); }
    void                        assign                      ( const char* str )                     { assign( str, strlen( str ) ); }
    void                        assign                      ( const char*, size_t );
    jstring                     get_jstring                 ()                                      { return (jstring)_jvalue.l; }

  //jstring                    _jstring;
};

//----------------------------------------------------------------------------Read_jstring_critical
// Nur kurzzeitig benutzen! Keine JNI-Aufrufe bis ~Read_jstring_critical!

struct Read_jstring_critical : Non_cloneable
{
                                Read_jstring_critical       ( JNIEnv*, jstring );
                               ~Read_jstring_critical       ();

    int                         length                      () const                                { return _jenv->GetStringLength( _jstring ); }
                                operator const jchar*       ()                                      { return _jchars; }
                              //operator const OLECHAR*     ()                                      { return (const OLECHAR*)_jchars; }
    const jchar&                operator []                 ( int i )                               { return _jchars[ i ]; }

    JNIEnv*                    _jenv;
    jstring                    _jstring;
    const jchar*               _jchars;
};

//--------------------------------------------------------------------------------------------Class

struct Class : Global_jobject  //Object, Has_vm
{
    Z_GNU_ONLY(                 Class                       ();  )                                  // Fr gcc 3.2. Nicht implementiert.
                                Class                       ( Vm* vm, const char* name = NULL );
                                Class                       ( Vm* vm, const string& name );

    void                        load                        ( const string& name );

    Class&                      operator =                  ( jclass c )                            { assign( c ); return *this; }
    jclass                      get_jclass                  ()                                      { return (jclass)get_jobject(); }
                                operator jclass             ()                                      { return (jclass)get_jobject(); }
/*
    jmethodID                   method_id                   ( const char*   name, const char*   signature ) { return env().get_method_id( (jclass)get_jobject(), name, signature ); }
    jmethodID                   method_id                   ( const string& name, const string& signature ) { return env().get_method_id( (jclass)get_jobject(), name, signature ); }

    jmethodID                   static_method_id            ( const char*   name, const char*   signature ) { return env().get_method_id( (jclass)get_jobject(), name, signature, true ); }
    jmethodID                   static_method_id            ( const string& name, const string& signature ) { return env().get_method_id( (jclass)get_jobject(), name, signature, true ); }
*/
  //jobject                     new_instance                ();
    jobject                     new_object                  ( const char* signature, ... );

    string                     _name;
};

//-------------------------------------------------------------------------------------------Method

struct Method
{
                                Method                      ( Class* cls, const char* name, const char* signature ) : _class(cls) { assign( name, signature ); }
                                Method                      ( Class* cls, const char* name_and_signature          ) : _class(cls) { assign( name_and_signature ); }

                                operator jmethodID          ()                                      { return id(); }

    void                        assign                      ( const char* name, const char* signature, bool is_static = false );
    void                        assign                      ( const string& name_and_signature, bool is_static = false );
    jmethodID                   id                          ()                                      { return _id; }


    Z_DEBUG_ONLY( string       _name_and_signature; )
    Class*                     _class;
    jmethodID                  _id;
};

//------------------------------------------------------------------------------------Static_method

struct Static_method : Method
{
                                Static_method               ( Class* cls, const char* name, const char* signature ) : Method( cls, "", "" ) { assign( name, signature ); }
                                Static_method               ( Class* cls, const char* name_and_signature ) : Method( cls, NULL ) { assign( name_and_signature ); }

    void                        assign                      ( const char* name, const char* signature )     { Method::assign( name, signature, true ); }
    void                        assign                      ( const string& name_and_signature )            { Method::assign( name_and_signature, true ); }
};

//---------------------------------------------------------------------------------------------Call
/*
struct Call
{
                                Call                        ( Env& e, jclass, const char* name_and_signature );
                                Call                        ( Env& e, jobject, const char* name_and_signature );

    void                        add_param                   ( jvalue v )                            { _jvalues[ _jvalues_count++ ] = v; }
    jobject                     call_unchecked_jobject      ();

    Fill_zero                  _zero_;
    jobject                    _jobject;
    jclass                     _jclass;
    jmethodID                  _method_id;
    int                        _is_static;
    int                        _jvalues_count;
    jvalue                     _jvalues [ max_params ];
};
*/
//------------------------------------------------------------------------------------inline Has_vm

inline                          Has_vm::Has_vm              ( Vm* vm )                              : _java_vm(vm) {}
inline void                     Has_vm::set_vm              ( Vm* vm )                              { _java_vm = vm; }
inline Vm*                      Has_vm::vm                  ()                                      { return _java_vm; }
inline JNIEnv*                  Has_vm::jenv                ()                                      { return _java_vm->jenv(); }
inline Env                      Has_vm::env                 ()                                      { return Env( _java_vm->jenv() ); }

//------------------------------------------------------------------------------------inline Jobject

inline                          Jobject::~Jobject           ()                                      {}  // Fr gcc 3.2

//-------------------------------------------------------------------------------------------------

} //namespace java
} //namespace zschimmer


#endif
