// $Id: z_posix.cxx 3774 2005-07-01 08:51:23Z jz $

#include "zschimmer.h"
#include "z_posix.h"
#include "mutex.h"


#ifdef Z_UNIX

#include <unistd.h>
#include <signal.h>
#include <math.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/socket.h>

#ifdef Z_SOLARIS
#   include <stropts.h>
#   include <sys/filio.h>       // ioctl( , FIONBIO )
#endif

using namespace std;
using namespace zschimmer;

//-------------------------------------------------------------------------------------------static

static Mutex                    interlocked_increment_mutex ( "InterlockedIncrement", Mutex::kind_nonrecursive );    // Fr InterlockedIncrement() und InterlockedDecrement()

//-----------------------------------------------------------------------------InterlockedIncrement

long32 InterlockedIncrement( long32* i )  
{ 
    long32 result = 0;

    Z_MUTEX( interlocked_increment_mutex )
    {
        (*i)++; 
        result = *i;
    }

    return result;
}

//-----------------------------------------------------------------------------InterlockedDecrement

long32 InterlockedDecrement( long32* i )  
{ 
    long result = 0;

    Z_MUTEX( interlocked_increment_mutex )
    {
        (*i)--; 
        result = *i; 
    }

    return result;
}

//-------------------------------------------------------------------------------------------------

namespace zschimmer {
namespace posix {

//-------------------------------------------------------------------------------------------------

void timespec_add( timespec* ts, double seconds )
{
    ts->tv_sec = (time_t)floor( seconds );
    ts->tv_nsec = (time_t)max( 0.0, ( seconds - floor( seconds ) ) * 1e9 );
}

//-------------------------------------------------------------------------------------------------

void timespec_add( timespec* ts, const timeval& tv )
{
    int u = ts->tv_nsec / 1000 + tv.tv_usec;
    ts->tv_nsec = u % 1000000 * 1000  + ts->tv_nsec % 1000;
    ts->tv_sec += tv.tv_sec + u / 1000000;
}

//-------------------------------------------------------------------------------------------------

} //namespace posix
} //namespace zschimmer

#endif
